{---------------------------------------------------------------------------}
{                                                                           }
{               P l a n e t s   C o m m a n d   C e n t e r                 }
{                                                                           }
{                             2003,2004 by Streu                            }
{                                                                           }
{---------------------------------------------------------------------------}
{                                                                           }
{                          FLAK Player / Simulator                          }
{                                                                           }
{---------------------------------------------------------------------------}
{$I switches.inc}
{.$DEFINE FLAK_DEBUG}
{$IFDEF FLAK_DEBUG}
  {$R+,Q+}
{$ENDIF}

{
  This module contains the FLAK battle engine, the visualisation engine,
  and the FLAK-related parts of the simulator.

  Naming conventions: similar to the C++ implementation.
     TFoo   - raw data, external representation, C++: 'TFoo'
     PFoo   - raw data pointer, C++: 'TFoo*'
     Foo    - runtime representation including status, C++: 'Foo'
     FooPtr - runtime representation pointer, C++: 'Foo*'
  Status is global here (as for the normal VCR), this frees us from
  having to carry a FlakBattle object around all the time.

  Since it is always known at the beginning of a fight how many ships
  will there be at most, the ship array (flak_ships) is pre-allocated,
  and it is an error to overflow. The fleet array (flak_fleets) and the
  attack lists (att_list) grow on demand.

  Rounding rules:
  - distances are NEVER rounded or truncated. All distance comparisons
    are exact, even if a round could have saved a cycle here or there.
  - coordinates are ALWAYS rounded arithmetically (i.e. to nearest integer,
    .5 rounds up).

  The simulator could be considered too deterministic because it uses
  the same RNG and seed to seed the FLAK player and to set up the
  battle, but the server has the same problem. Actually, the server
  even uses only a 16 bit seed currently.

  Currently, the code seems to be very sensitive to floating point
  precision. The original server implementation was troubled by the
  famous gcc bug (implicitly converts some 'double' variables to 'long
  double'). This implementation, using 48 bit software FP, yields the
  same results as the FLAK server on x86 using 64 bit FP, *except* that
  'Trunc(damage)' occasionally differs by one. The test case was a
  Ragnarok style fight with 504 ships, yielding a 70 kB flak.hst and
  a 400 MB trace file.
}

UNIT Flak;

INTERFACE

USES PConfig, Objects;

TYPE PFlakShip = ^TFlakShip;
     TFlakShip = RECORD
                    name                  : ARRAY[1..20] OF CHAR;
                    damage_init           : INTEGER;
                    crew_init             : INTEGER;
                    id                    : INTEGER;
                    player                : INTEGER;
                    hull                  : INTEGER;
                    level                 : INTEGER;
                    beam_count            : INTEGER;
                    beam_type             : INTEGER;
                    torp_lcount           : INTEGER;
                    torp_count_init       : INTEGER;
                    torp_type             : INTEGER;
                    bay_count             : INTEGER;
                    fighter_count_init    : INTEGER;
                    mass                  : INTEGER;
                    shield_init           : INTEGER;
                    max_fighters_launched : INTEGER;
                    rating                : LONGINT;
                    compensation_rating   : INTEGER;
                    flags                 : INTEGER;
                    ending_status         : INTEGER;
                 END;

CONST flak_IsPlanet = 1;

TYPE PFlakFleet = ^TFlakFleet;
     TFlakFleet = RECORD
                     player           : INTEGER;
                     first_ship       : INTEGER;
                     num_ships        : INTEGER;
                     speed            : INTEGER;
                     att_list_pointer : LONGINT;
                     att_list_size    : LONGINT;
                     x_init, y_init   : LONGINT;
                  END;

TYPE PFlakBattle = ^TFlakBattle;
     TFlakBattle =  RECORD
                       this_size     : LONGINT;
                       x, y          : INTEGER;
                       seed          : LONGINT;
                       total_time    : LONGINT;
                       ambient_flags : LONGINT;
                       stuff         : ARRAY[(iFleets,iShips,iAttList)] OF RECORD
                                          num  : LONGINT;
                                          size : LONGINT;
                                          pos  : LONGINT;
                                       END;
                    END;

TYPE TFlakHeader = RECORD
                      magic           : ARRAY[1..8] OF CHAR;
                      filefmt_version : INTEGER;
                      player          : INTEGER;
                      turn            : INTEGER;
                      num_battles     : INTEGER;
                      timestamp       : ARRAY[1..18] OF CHAR;
                      reserved        : LONGINT;
                   END;

{ Manifest constants of the algorithm }
CONST
      { We choose a new enemy every this many battle ticks }
      CHOOSE_ENEMY_TIME                = 30;

      { Torps move at this velocity. Fighter and ship speeds are set in
        the pconfig }
      TORP_MOVEMENT_SPEED              = 1000;

      { Rating difference is increased by this value before comparison }
      DIFF_OFFSET                      = 100;

      { Rating is divided by a divisor. Divisor starts at one of these
        values, depending upon whether the target is bigger or smaller... }
      DIVISOR_IF_SMALLER               = 100;
      DIVISOR_IF_BIGGER                = 150;
      { ...plus this if we're currently firing at that unit... }
      DIVISOR_SAME_ENEMY_BONUS         = 50;

      { The multiplier equals the distance, but at least this value. }
      MULTIPLIER_MIN                   = 50;

      { Distance at which we can do fighter intercept. PVCR has 128 here. }
      FIGHTER_INTERCEPT_RANGE          = 128;

      { If set, we do fighter intercept }
      FIGHTER_INTERCEPT                = TRUE;

      { If set, anti-fighter beams consider fleets, not ships }
      ANTIFIGHTER_BEAMS_USE_FLEETS     = TRUE;

      { If set, MaxFightersLaunched is dynamically related to bays }
      DYNAMIC_MFL_LIMIT                = TRUE;

      { Minimum MaxFightersLaunched per ship }
      MINIMUM_MFL                      = 1;
      MAXIMUM_MFL                      = 50;
      MFL_SCALE                        = 2;

      { Divisor for compensation formula. Should remain fixed. }
      COMPENSATION_DIVISOR             = 1000;

      { Maximum compensation bonus }
      COMPENSATION_LIMIT               = 2;

      ALTERNATE_COMPENSATION           = FALSE;

      FLAK_MAX_BEAMS = 20;
      FLAK_MAX_TORPS = 20;
      FLAK_MAX_BAYS = 50;
      FLAK_MAX_PLAYERS = 12;

TYPE FlakShip = RECORD
                  { running status }
                  status :  RECORD
                              shield, damage, crew : REAL;
                              last_hit_by          : INTEGER; { ship Id or -1 }
                              is_alive             : BOOLEAN;
                              BeamStatus           : ARRAY[1..FLAK_MAX_BEAMS] OF INTEGER;
                              TorpStatus           : ARRAY[1..FLAK_MAX_TORPS] OF INTEGER;
                              BayStatus            : ARRAY[1..FLAK_MAX_BAYS] OF INTEGER;
                              torp_limit           : INTEGER;
                              LaunchCountdown      : INTEGER;
                              fighters_launched    : INTEGER;
                              z_offs               : INTEGER; { FIXME: needed? }
                              fighter_count        : INTEGER;
                              torp_count           : INTEGER;
                              number               : INTEGER; { my Id (index into flak_ships) }
                            END;
                  { PConfig }
                  config : TPBattleConfig;
                  { original ship data }
                  data   : TFlakShip;
                  { statistics for simulator }
                  stat   : RECORD
                             min_ftr_here : INTEGER;
                             torps_hit    : INTEGER;
                           END;
                  { number of the fleet I'm in }
                  fleet_nr : INTEGER;
                END;
     FlakShipPtr = ^FlakShip;

FUNCTION  FlakInit(max : INTEGER) : BOOLEAN;
PROCEDURE FlakDone;
FUNCTION  FlakAdd(CONST item : TFlakShip; aux : POINTER) : BOOLEAN;
FUNCTION  FlakInitSimulation : BOOLEAN;
PROCEDURE FlakInitCommon(VAR item : TFlakShip);
FUNCTION  FlakEndCheck:BOOLEAN;
FUNCTION  FlakPlayOneTick : BOOLEAN;
PROCEDURE FlakBattleInit;
PROCEDURE FlakPlayVis(auto_figure_result:BOOLEAN);
FUNCTION  FlakGetShipByAux(aux : POINTER) : FlakShipPtr;
PROCEDURE FlakFigureResults;
PROCEDURE FlakPlayFile(VAR t : TStream; CONST s : STRING; current : POINTER);

VAR
  flak_rng_seed    : LONGINT;
  flak_time        : LONGINT;
  flak_header_seed : LONGINT;
  flak_header_time : LONGINT;
  flak_oom         : BOOLEAN;

CONST
  vis_inhibit : BOOLEAN = FALSE;

IMPLEMENTATION

USES UI, Drivers, Config, HConfig, Global, ccSim, LowLevel, Hullfunc,
     FCode, MsgWin, User, Swapper, VGARegs, AAFont, Keyboard, ccHelp;

TYPE PIntArray = ^TIntArray;
     TIntArray = ARRAY[0..1000] OF INTEGER;
     PInteger = ^INTEGER;

{ FlakObject: a torpedo or fighter in free space. }
TYPE
  FlakObjectKind = (oFighter, oTorpedo, oDeleteMe);
  FlakObjectPtr     = ^FlakObject;
  FlakObject     = RECORD
                      kind                   : FlakObjectKind;
                      can_change_enemy       : BOOLEAN;
                      x, y                   : LONGINT;
                      z                      : INTEGER;
                      enemy                  : INTEGER;
                      owner                  : INTEGER;
                      strikes                : INTEGER;
                      kill, expl, death_flag : INTEGER;
                      speed                  : INTEGER;
                      next                   : FlakObjectPtr;
                    END;

{ Attack lists. }
TYPE TAttackListItem = RECORD
                         ship_nr  : INTEGER;
                         divi_add : INTEGER;
                       END;
     TAttackList     = ARRAY[0 .. 65520 DIV 4] OF TAttackListItem;
     PAttackList     = ^TAttackList;

{ Fleet runtime representation. }
TYPE FlakFleet = RECORD
                   data         : TFlakFleet;
                   enemy        : INTEGER;
                   alive        : BOOLEAN;
                   x, y         : LONGINT;
                   new_x, new_y : LONGINT;
                 END;
     FlakFleetPtr = ^FlakFleet;

{ Ship (runtime) array }
TYPE PFlakShipArray = ^TFlakShipArray;
     TFlakShipArray = ARRAY[0..999] OF RECORD
       fs  : FlakShipPtr;
       aux : PSimObject;
     END;

{ Fleet (runtime) array }
TYPE PFlakFleetArray = ^TFlakFleetArray;
     TFlakFleetArray = ARRAY[0..999] OF FlakFleetPtr;

{ Player-related status }
TYPE FlakPlayer = RECORD
                     number         : INTEGER;
                     num_live_ships : INTEGER;
                     sum_strength   : LONGINT;
                     stuff          : FlakObjectPtr;
                   END;

VAR flak_ships       : PFlakShipArray;
    flak_ship_cnt    : INTEGER;
    flak_ship_alloc  : INTEGER;

    flak_fleets      : PFlakFleetArray;
    flak_fleet_cnt   : INTEGER;
    flak_fleet_alloc : INTEGER;

    att_list        : PAttackList;
    att_list_cnt    : INTEGER;
    att_list_alloc  : INTEGER;

    players         : ARRAY[1..FLAK_MAX_PLAYERS] OF FlakPlayer;

PROCEDURE VisDrawBeam(x1,y1:LONGINT; z1:INTEGER; x2,y2:LONGINT; z2:INTEGER); FORWARD;

{ Our random-number generator. Probably this one is equivalent to the
  Turbo Pascal 'random' function. }
FUNCTION FlakRandom(limit : INTEGER) : INTEGER;
BEGIN
  {$R-,Q-}
  flak_rng_seed := $8088405 * flak_rng_seed + 1;
  FlakRandom := ((flak_rng_seed SHR 16) * limit) SHR 16;
  {$IFDEF FLAK_DEBUG}{$R+,Q+}{$ENDIF}
END;

{ Initialize for a fight with at most /max/ ships }
FUNCTION FlakInit(max : INTEGER) : BOOLEAN;
BEGIN
  att_list_cnt := 0;
  att_list_alloc := 0;
  att_list := NIL;
  flak_fleets := NIL;
  flak_fleet_cnt := 0;
  flak_fleet_alloc := 0;
  flak_ships := Malloc(Sizeof(flak_ships^[0]) * max);
  IF flak_ships<>NIL THEN BEGIN
    flak_ship_cnt := 0;
    flak_ship_alloc := max;
    FlakInit := TRUE;
  END ELSE BEGIN
    flak_oom := TRUE;
    FlakInit := FALSE;
  END;
END; { FlakInit }

{ Exit }
PROCEDURE FlakDone;
VAR i : INTEGER;
BEGIN
  IF flak_ships<>NIL THEN BEGIN
    FOR i:=0 TO flak_ship_alloc-1 DO
      IF flak_ships^[i].fs<>NIL THEN
        Dispose(flak_ships^[i].fs);
    FreeMem(flak_ships, flak_ship_alloc * Sizeof(flak_ships^[0]));
  END;
  flak_ships := NIL;
  flak_ship_cnt := 0;
  flak_ship_alloc := 0;

  IF flak_fleets<>NIL THEN BEGIN
    FOR i:=0 TO flak_fleet_alloc-1 DO
      IF flak_fleets^[i]<>NIL THEN
        Dispose(flak_fleets^[i]);
    FreeMem(flak_fleets, flak_fleet_alloc * Sizeof(flak_fleets^[0]));
  END;
  flak_fleets := NIL;
  flak_fleet_cnt := 0;
  flak_fleet_alloc := 0;

  IF att_list<>NIL THEN
   FreeMem(att_list, att_list_alloc * Sizeof(att_list^[0]));
  att_list := NIL;
  att_list_alloc := 0;
  att_list_cnt := 0;
END; { FlakDone }

{ Add a ship to the simulation. 'item' is the normal data, 'aux' is
 an auxiliary pointer the caller may use. This is used to store the
 backlink to the PSimObject for the simulator. }
FUNCTION FlakAdd(CONST item : TFlakShip; aux : POINTER{PSimObject}):BOOLEAN;
VAR p : FlakShipPtr;
BEGIN
  IF flak_ship_cnt >= flak_ship_alloc THEN BEGIN
    FlakAdd := FALSE
  END ELSE BEGIN
    p := Malloc(Sizeof(FlakShip));
    IF p<>NIL THEN BEGIN
      p^.data := item;
      flak_ships^[flak_ship_cnt].fs := p;
      flak_ships^[flak_ship_cnt].aux := aux;
      Inc(flak_ship_cnt);
      FlakAdd := TRUE;
    END ELSE BEGIN
      flak_oom := TRUE;
      FlakAdd := FALSE;
    END;
  END;
END; { FlakAdd }

{ Get a ship with a particular aux pointer. Returns 0 if no such ship. }
FUNCTION FlakGetShipByAux(aux : POINTER) : FlakShipPtr;
VAR i : INTEGER;
BEGIN
  IF aux<>NIL THEN
    FOR i := 0 TO flak_ship_cnt-1 DO
      IF flak_ships^[i].aux = aux THEN BEGIN
        FlakGetShipByAux := flak_ships^[i].fs;
        Exit;
      END;
  FlakGetShipByAux := NIL;
END;

{ Given a ship /s/ with only the data filled in, compute the other fields. }
PROCEDURE FlakShipInit(VAR s : FlakShip; number : INTEGER; fleet_nr:INTEGER);

  { Update config struct /c/ for experience /level/ }
  PROCEDURE AddExperience(VAR c : TPBattleConfig; level : INTEGER);
  BEGIN
    IF level > pconf^.main.NumExperienceLevels THEN level:=pconf^.main.NumExperienceLevels;
    IF level > 0 THEN BEGIN
      Inc(c.BayRechargeRate,      pconf^.main.EModBayRechargeRate[level]);
      Inc(c.BayRechargeBonus,     pconf^.main.EModBayRechargeBonus[level]);
      Inc(c.BeamRechargeRate,     pconf^.main.EModBeamRechargeRate[level]);
      Inc(c.BeamRechargeBonus,    pconf^.main.EModBeamRechargeBonus[level]);
      Inc(c.TubeRechargeRate,     pconf^.main.EModTubeRechargeRate[level]);
      Inc(c.BeamHitFighterCharge, pconf^.main.EModBeamHitFighterCharge[level]);
      Inc(c.TorpHitOdds,          pconf^.main.EModTorpHitOdds[level]);
      Inc(c.BeamHitOdds,          pconf^.main.EModBeamHitOdds[level]);
      Inc(c.BeamHitBonus,         pconf^.main.EModBeamHitBonus[level]);
      Inc(c.StrikesPerFighter,    pconf^.main.EModStrikesPerFighter[level]);
      Inc(c.FighterBeamExplosive, pconf^.main.EModFighterBeamExplosive[level]);
      Inc(c.FighterBeamKill,      pconf^.main.EModFighterBeamKill[level]);
      Inc(c.FighterMovementSpeed, pconf^.main.EModFighterMovementSpeed[level]);
      Inc(c.MaxFightersLaunched,  pconf^.main.EModMaxFightersLaunched[level]);
      Inc(c.TorpHitBonus,         pconf^.main.EModTorpHitBonus[level]);
      Inc(c.TubeRechargeBonus,    pconf^.main.EModTubeRechargeBonus[level]);
      Inc(c.ShieldKillScaling,    pconf^.main.EModShieldKillScaling[level]);
      Inc(c.ShieldDamageScaling,  pconf^.main.EModShieldDamageScaling[level]);
      Inc(c.CrewKillScaling,      pconf^.main.EModCrewKillScaling[level]);
      Inc(c.HullDamageScaling,    pconf^.main.EModHullDamageScaling[level]);

      IF c.MaxFightersLaunched > PMaxFighters THEN
       c.MaxFightersLaunched := PMaxFighters;
    END;
  END;

  { The following functions were stolen from ccVCR }
  FUNCTION P_BayRechargeRate(num:INTEGER; CONST fig:TPBattleConfig):INTEGER;
  VAR i:INTEGER;
  BEGIN
    i := fig.BayRechargeBonus * num + fig.BayRechargeRate;
    IF i<0 THEN P_BayRechargeRate := 0 ELSE P_BayRechargeRate := i;
  END;

  FUNCTION P_BeamHitOdds(beam:INTEGER; CONST fig:TPBattleConfig):INTEGER;
  VAR i:INTEGER;
  BEGIN
    i:=(fig.BeamHitBonus * (Beams[beam].Kill + Beams[beam].Damage)) DIV 100
       + fig.BeamHitOdds;
    IF i<0 THEN P_BeamHitOdds:=0 ELSE P_BeamHitOdds:=i;
  END;

  FUNCTION P_BeamRechargeRate(beam:INTEGER; CONST fig:TPBattleConfig):INTEGER;
  VAR i:INTEGER;
  BEGIN
    i := ((Beams[beam].Kill + Beams[beam].Damage) * fig.BeamRechargeBonus) DIV 100
         + fig.BeamRechargeRate;
    IF i<0 THEN P_BeamRechargeRate:=0 ELSE P_BeamRechargeRate:=i;
  END;

  FUNCTION P_TorpHitOdds(torp:INTEGER; CONST fig:TPBattleConfig):INTEGER;
  VAR i:INTEGER;
  BEGIN
    i:=(fig.TorpHitBonus * (Torps[torp].Kill + Torps[torp].Damage)) DIV 100
       + fig.TorpHitOdds;
    IF i<0 THEN P_TorpHitOdds:=0 ELSE P_TorpHitOdds:=i;
  END;

  FUNCTION P_TubeRechargeRate(torp:INTEGER; CONST fig:TPBattleConfig):INTEGER;
  VAR i:INTEGER;
  BEGIN
    i := ((Torps[torp].Kill + Torps[torp].Damage) * fig.TubeRechargeBonus) DIV 100
         + fig.TubeRechargeRate;
    IF i<0 THEN P_TubeRechargeRate:=0 ELSE P_TubeRechargeRate:=i;
  END;

VAR i, st :  INTEGER;
BEGIN
  s.status.shield            := s.data.shield_init;
  s.status.damage            := s.data.damage_init;
  s.status.crew              := s.data.crew_init;
  s.status.is_alive          := TRUE;
  s.status.last_hit_by       := -1;
  s.status.torp_limit        := s.data.torp_lcount;
  s.status.LaunchCountdown  := 0;
  s.status.fighters_launched := 0;
  s.status.z_offs            := 0;   { will be computed elsewhere }
  s.status.fighter_count     := s.data.fighter_count_init;
  s.status.torp_count        := s.data.torp_count_init;
  s.status.number            := number;

  IF (s.data.player <= 0) OR (s.data.player > 11) THEN i := 11 ELSE i := s.data.player;
  s.config := pbattleconf^[i];
  AddExperience(s.config, s.data.level);

  { Now compute the effective rates }
  s.config.BayRechargeRate := P_BayRechargeRate(s.data.bay_count, s.config);
  IF s.data.beam_count>0 THEN BEGIN
    s.config.BeamRechargeRate := P_BeamRechargeRate(s.data.beam_type, s.config);
    s.config.BeamHitOdds := P_BeamHitOdds(s.data.beam_type, s.config);
  END;
  IF s.data.torp_lcount>0 THEN BEGIN
    s.config.TubeRechargeRate := P_TubeRechargeRate(s.data.torp_type, s.config);
    s.config.TorpHitOdds := P_TorpHitOdds(s.data.torp_type, s.config);
  END;

  IF s.data.shield_init = 100 THEN st := 1000 ELSE st := 0;
  FOR i:=1 TO FLAK_MAX_BEAMS DO s.status.BeamStatus[i] := st;
  FOR i:=1 TO FLAK_MAX_TORPS DO s.status.TorpStatus[i] := st;
  FOR i:=1 TO FLAK_MAX_BAYS  DO s.status.BayStatus[i] := st;

  s.stat.min_ftr_here := s.status.fighter_count;
  s.stat.torps_hit := 0;

  s.fleet_nr := fleet_nr;
END;

{ Recharge }
PROCEDURE FlakShipRecharge(VAR s : FlakShip);
VAR i : INTEGER;
BEGIN
  IF NOT s.status.is_alive THEN Exit;
  IF s.status.LaunchCountdown > 0 THEN Dec(s.status.LaunchCountdown);
  FOR i:=1 TO s.data.bay_count DO
    IF s.status.BayStatus[i] < 1000 THEN
      Inc(s.status.BayStatus[i], FlakRandom(s.config.BayRechargeRate));
  FOR i:=1 TO s.data.beam_count DO
    IF s.status.BeamStatus[i] < 1000 THEN
      Inc(s.status.BeamStatus[i], FlakRandom(s.config.BeamRechargeRate));
  FOR i:=1 TO s.data.torp_lcount DO
    IF s.status.TorpStatus[i] < 1000 THEN
      Inc(s.status.TorpStatus[i], FlakRandom(s.config.TubeRechargeRate));
END;

{********* FlakPlayer **********}

{ Garbage-collect player pl's space stuff (torps/fighters) }
PROCEDURE FlakPlayerGC(VAR pl : FlakPlayer);
VAR
  p  : FlakObjectPtr;
  pp : ^FlakObjectPtr;
BEGIN
  pp := @pl.stuff;
  WHILE pp^ <> NIL DO BEGIN
    p := pp^;
    IF p^.kind = oDeleteMe THEN BEGIN
      pp^ := p^.next;
      Dispose(p);
    END ELSE pp := @p^.next;
  END;
END;

{ Initialize a player }
PROCEDURE FlakPlayerInit(number : INTEGER);
BEGIN
  players[number].number := number;
  players[number].num_live_ships := 0;
  players[number].sum_strength := 0;
  players[number].stuff := NIL;
END;

{ Delete a player }
PROCEDURE FlakPlayerDone(number : INTEGER);
VAR p, q : FlakObjectPtr;
BEGIN
  p := players[number].stuff;
  players[number].stuff := NIL;
  WHILE p<>NIL DO BEGIN
    q := p;
    p := p^.next;
    Dispose(q);
  END;
END; { FlakPlayerDone }

{********* FlakObject ********}

{ Move object /p/ towards /x,y/ by at most /amt/ meters. Returns 0
  if waypoint reached. }
FUNCTION FlakObjectMoveTowards(VAR o : FlakObject; x, y:LONGINT; z:INTEGER; amt:INTEGER):LONGINT;
VAR dist : REAL;
    nd   : REAL;
BEGIN
  dist := Sqrt(Sqr(0.0 + x - o.x) + Sqr(0.0 + y - o.y));
  nd := dist - amt;
  IF nd <= 0 THEN BEGIN
    o.x := x;
    o.y := y;
    o.z := z;
    FlakObjectMoveTowards := 0;
  END ELSE BEGIN
    {$IFOPT N+}'Port me. This must be real arithmetic round, not IEEE to-nearest-or-even.'{$ENDIF}
    o.x := x + Round((o.x - x) * nd / dist);
    o.y := y + Round((o.y - y) * nd / dist);
    o.z := z + Round((o.z - z) * nd / dist);
    FlakObjectMoveTowards := Trunc(nd);
  END;
END;

{********** FlakFleet ********}

PROCEDURE FlakFleetInit(VAR f : FlakFleet);
BEGIN
  f.alive := TRUE;
  f.enemy := -1;
  f.x := f.data.x_init;
  f.y := f.data.y_init;
  f.new_x := 0;
  f.new_y := 0;
END; { FlakFleetInit }

{ Add a fleet. Returns true on success. }
FUNCTION FlakFleetAdd : BOOLEAN;
VAR p : PFlakFleetArray;
    i : INTEGER;
BEGIN
  IF (flak_fleet_cnt >= flak_fleet_alloc) THEN BEGIN
    i := flak_fleet_alloc + 32;
    p := Realloc(Sizeof(flak_fleets^[0]) * i,
                 flak_fleets,
                 Sizeof(flak_fleets^[0]) * flak_fleet_alloc);
    IF p=NIL THEN BEGIN
      flak_oom := TRUE;
      FlakFleetAdd := FALSE;
      Exit;
    END;
    flak_fleets := p;
    flak_fleet_alloc := i;
  END;

  flak_fleets^[flak_fleet_cnt] := Malloc(Sizeof(FlakFleet));
  IF flak_fleets^[flak_fleet_cnt] = NIL THEN BEGIN
    flak_oom := TRUE;
    FlakFleetAdd := FALSE;
    Exit;
  END;
  Inc(flak_fleet_cnt);
  FlakFleetAdd := TRUE;
END; { FlakFleetAdd }

{ Add attack list entry. Returns true on success. }
FUNCTION FlakAddAttListEntry(ship, bonus : INTEGER):BOOLEAN;
VAR p : PAttackList;
    i : INTEGER;
BEGIN
  IF (att_list_cnt >= att_list_alloc) THEN BEGIN
    i := att_list_alloc + 64;
    IF i > (65500 DIV Sizeof(TAttackListItem)) THEN BEGIN
      flak_oom := TRUE;
      FlakAddAttListEntry := FALSE;
      Exit;
    END;

    p := Realloc(Sizeof(TAttackListItem) * i,
                 att_list,
                 Sizeof(TAttackListItem) * att_list_alloc);
    IF p=NIL THEN BEGIN
      flak_oom := TRUE;
      FlakAddAttListEntry := FALSE;
      Exit;
    END;
    att_list := p;
    att_list_alloc := i;
  END;

  att_list^[att_list_cnt].ship_nr := ship;
  att_list^[att_list_cnt].divi_add := bonus;
  Inc(att_list_cnt);
  FlakAddAttListEntry := TRUE;
END; { FlakAddAttListEntry }

{********** FlakBattle *********}

{ These functions taken from ccVCR }
FUNCTION P_ShieldDamage(Expl, Kill, Mass:INTEGER; VAR Who:FlakShip):REAL;
VAR Damage:REAL;
BEGIN
  Damage := (who.config.ShieldDamageScaling * LONGINT(Expl)
           + who.config.ShieldKillScaling * LONGINT(Kill)) / (Mass+1);
  IF (Damage>10000) THEN Damage:=10000;
  IF who.config.AlternativeCombat=0
   THEN P_ShieldDamage := Trunc(Damage + 1.5)
   ELSE P_ShieldDamage := Damage;
END;

FUNCTION P_HullDamage(Expl, Kill, Mass:INTEGER; VAR Who:FlakShip):REAL;
VAR d:REAL;
BEGIN
  IF who.config.AlternativeCombat<>0 THEN BEGIN
    d := Expl * (who.config.HullDamageScaling / Succ(Mass));
    IF d>10000 THEN d:=10000;
    P_HullDamage := d;
  END ELSE BEGIN
    d := P_ShieldDamage(Expl, Kill, Mass, Who) * who.config.HullDamageScaling / Succ(Mass);
    IF d>10000 THEN d:=10000;
    P_HullDamage := Trunc(d+1.5);
  END;
END;

FUNCTION P_CrewKilled(Kill, Mass, DRFlag:INTEGER; VAR Who:FlakShip):REAL;
VAR d, l:REAL;
BEGIN
  d := who.config.CrewKillScaling * (Kill / Succ(Mass));
  IF who.config.AlternativeCombat<>0 THEN P_CrewKilled := d
  ELSE BEGIN
    l := Trunc(d + 0.5);
    IF (DRFlag=0) AND (l=0) THEN P_CrewKilled := 1 ELSE P_CrewKilled := l;
  END;
END;

{ Distance between two fleets }
FUNCTION DistanceFleetToFleet(VAR a, b : FlakFleet) : REAL;
BEGIN
  DistanceFleetToFleet := Sqrt(Sqr(a.x - b.x + 0.0) + Sqr(a.y - b.y + 0.0));
END;

{ Distance between fleet and /x,y/ }
FUNCTION DistanceFleetToXY(x, y:LONGINT; CONST a : FlakFleet) : REAL;
BEGIN
  DistanceFleetToXY := Sqrt(Sqr(a.x - x + 0.0) + Sqr(a.y - y + 0.0));
END;

{ Get fleet of a ship }
FUNCTION FleetOf(ship_nr:INTEGER):FlakFleetPtr;
BEGIN
  FleetOf := flak_fleets^[flak_ships^[ship_nr].fs^.fleet_nr];
END;

{ Distance between fleet and ship }
FUNCTION DistanceFleetToShip(VAR f : FlakFleet; ship_nr : INTEGER) : REAL;
BEGIN
  DistanceFleetToShip := DistanceFleetToFleet(f, FleetOf(ship_nr)^);
END; { DistanceFleetToShip }

{ Compare distance using no FP arithmetic }
FUNCTION IsDistanceLERadius(x1,y1,x2,y2:LONGINT; r : LONGINT) : BOOLEAN;
VAR dx, dy : LONGINT;
BEGIN
  dx := Abs(x2-x1);
  dy := Abs(y2-y1);
  IsDistanceLERadius := (dx <= r) AND (dy <= r) AND (Sqr(dx)+Sqr(dy) <= Sqr(r));
END;

{ Compute torpedo limit. We do not use more torpedoes than needed to
  crush the enemy }
PROCEDURE ComputeTorpLimit(VAR s, enemy : FlakShip; divisor : INTEGER);
VAR hd, sd, cd, v1, v2 : REAL;
    expl, kill         : INTEGER;
    limit              : INTEGER;
    torps_reqd         : INTEGER;
BEGIN
  s.status.torp_limit := s.data.torp_lcount;
  IF s.data.torp_lcount <= 2 THEN Exit;      { don't bother for small ships }
  IF s.config.TorpHitOdds <= 0 THEN Exit;  { pathological case }

  expl := Torps[s.data.torp_type].Damage;
  kill := Torps[s.data.torp_type].Kill;
  IF s.config.AlternativeCombat=0 THEN BEGIN
    Inc(expl, expl);
    Inc(kill, kill);
  END;


  cd := P_CrewKilled(kill, enemy.data.mass, expl, enemy);
  IF expl=0 THEN BEGIN                  { death ray }
    IF enemy.data.flags AND flak_IsPlanet<>0 THEN Exit;       { should not happen }
    torps_reqd := Trunc(1 + (enemy.status.crew / cd) * 150 / s.config.TorpHitOdds);
  END ELSE BEGIN
    hd := P_HullDamage(expl, kill, enemy.data.mass, enemy);
    sd := P_ShieldDamage(expl, kill, enemy.data.mass, enemy);

    IF RaceIds[enemy.data.player]=2 THEN limit:=150 ELSE limit:=100;
    v1 := (limit-enemy.status.damage) / (hd + 0.01);
    v2 := enemy.status.crew / (cd + 0.01);
    IF (v1 > v2) THEN v1 := v2;

    torps_reqd := Trunc(1 + (enemy.status.shield / sd + v1) * 150 / s.config.TorpHitOdds);
  END;

  IF divisor > 0 THEN
    torps_reqd := (torps_reqd + divisor-1) DIV divisor;

  IF torps_reqd < s.data.torp_lcount THEN
    s.status.torp_limit := torps_reqd;
END; { ComputeTorpLimit }

{ Hit ship /sh/ with weapon /expl,kill,deathflag/. expl and kill come
  from the weapon specs, deathflag is zero if it is a death ray. This is
  almost the same as in PVCR, but we don't interrupt immediately when
  the ship is dead. }
PROCEDURE HitShipWith(ship_id : INTEGER; firing_ship : INTEGER; expl, kill, deathflag : INTEGER);
VAR damage_rate : REAL;
    damage      : REAL;
    sh, ene     : FlakShipPtr;
    opp_strength, att_strength : LONGINT;
    opp_count, att_count       : INTEGER;
    scale       : REAL;
    eff_mass    : INTEGER;
    adder       : INTEGER;

LABEL Vis, DeathRay;
BEGIN
  sh := flak_ships^[ship_id].fs;
  ene := flak_ships^[firing_ship].fs;

  { compute strength correction factor }
  att_count := players[ene^.data.player].num_live_ships;
  opp_count := players[sh^.data.player].num_live_ships;
  att_strength := players[ene^.data.player].sum_strength;
  opp_strength := players[sh^.data.player].sum_strength;

  IF (att_count < opp_count) AND (att_strength < opp_strength) THEN BEGIN
    IF (att_strength+COMPENSATION_DIVISOR) * COMPENSATION_LIMIT < (opp_strength+COMPENSATION_DIVISOR)
     THEN eff_mass := sh^.data.mass DIV COMPENSATION_LIMIT
     ELSE eff_mass := sh^.data.mass
                     * LONGINT(att_strength+COMPENSATION_DIVISOR) DIV LONGINT(opp_strength+COMPENSATION_DIVISOR);
  END ELSE BEGIN
    eff_mass := sh^.data.mass;
  END;

  { normal damage processing }
  damage_rate := 1.0;
  IF Kill<=0 THEN Kill:=1;
  IF Expl<=0 THEN Expl:=1;
  IF (DeathFlag=0) THEN GOTO DeathRay;
  IF sh^.status.shield>0 THEN BEGIN
    damage := P_ShieldDamage(expl, kill, eff_mass, sh^);
    IF sh^.status.shield <= damage THEN BEGIN
      {... shield completely down ...}
      damage_rate := (damage-sh^.status.shield)/damage;
      sh^.status.shield := 0;
    END ELSE BEGIN
      damage_rate := 0;  { shields still there, no damage }
      sh^.status.shield := sh^.status.shield - damage;
    END;
  END;
  IF damage_rate <= 0 THEN GOTO Vis;

  damage := P_HullDamage(Expl, Kill, eff_mass, sh^) * damage_rate;
  sh^.status.damage := sh^.status.damage + damage;
  IF sh^.status.damage > 9999 THEN sh^.status.damage := 9999;

 DeathRay:
  IF (sh^.data.flags AND flak_IsPlanet) = 0 THEN BEGIN
    damage := P_CrewKilled(Kill, eff_mass, DeathFlag, sh^) * damage_rate;
    sh^.status.crew := sh^.status.crew - damage;
    IF sh^.status.crew < 0.5 THEN sh^.status.crew:=0;
  END;

 Vis:
  { we'd display nice animations here }
  IF sh^.status.is_alive THEN sh^.status.last_hit_by := firing_ship;
END;

{ Check whether /we/ can attack /they/. }
FUNCTION CanStillAttack(VAR we, they : FlakShip):BOOLEAN;
VAR torpc, beamc : INTEGER;
BEGIN
  torpc := we.data.torp_lcount;
  beamc := we.data.beam_count;

  IF they.data.flags AND flak_IsPlanet<>0 THEN BEGIN
    { discount death rays against planet }
    IF (torpc<>0) AND (Torps[we.data.torp_type].damage = 0) THEN torpc := 0;
    IF (beamc<>0) AND (Beams[we.data.beam_type].damage = 0)
      AND (they.status.fighter_count = 0)
      AND (they.status.fighters_launched = 0)
     THEN beamc := 0;
  END;

  CanStillAttack := ((beamc<>0)
                     OR ((torpc<>0) AND (we.status.torp_count <> 0))
                     OR ((we.data.bay_count <> 0)
                         AND ((we.status.fighter_count <> 0)
                              OR (we.status.fighters_launched <> 0))));
END; { CanStillAttack }

{ Choose enemy for a fleet }
PROCEDURE ChooseEnemy(VAR f : FlakFleet);
VAR i             : INTEGER;
    index         : INTEGER;
    attack_rating : LONGINT;
    their_rating  : LONGINT;
    n             : INTEGER;
    we            : FlakShipPtr;
    their_ship    : FlakShipPtr;
    diff          : LONGINT;
    divisor       : INTEGER;
    multiplier    : LONGINT;
    num_torpers   : INTEGER;
    torpc, beamc  : INTEGER;

    best          : INTEGER;
    best_diff     : LONGINT;
    eff_rating_bonus, this_rating_bonus : INTEGER;

BEGIN
  best := -1;
  best_diff := MAXLONGINT;

  { is this fleet already dead? }
  IF NOT f.alive THEN Exit;

  { for each ship, compute a rating; pick best one. }
  FOR index := 0 TO f.data.att_list_size-1 DO BEGIN
    i := att_list^[index + f.data.att_list_pointer].ship_nr;
    { Quick exit if it is ours or dead }
    their_ship := flak_ships^[i].fs;
    IF their_ship^.data.player = f.data.player THEN Continue;
    IF NOT their_ship^.status.is_alive THEN Continue;

    { Now compute our attack rating. A ship is only considered if it's not
      empty. The problem case
         ship with death rays  +  ship with torps (none left)
      against
         planet
      needs to be handled specially.
      FIXME: this is not yet implemented. }
    attack_rating := 0;
    eff_rating_bonus := att_list^[index + f.data.att_list_pointer].divi_add;
    FOR n:=1 TO f.data.num_ships DO BEGIN
      we := flak_ships^[f.data.first_ship + (n-1)].fs;
      IF we^.status.is_alive AND (CanStillAttack(we^, their_ship^) OR CanStillAttack(their_ship^, we^))
      THEN Inc(attack_rating, we^.data.rating);
    END;

    { Attack rating is the rating which we have against the enemy.
      If it is zero, we probably can't attack it. }
    IF attack_rating = 0 THEN Continue;

    their_rating := their_ship^.data.rating;

    IF attack_rating < their_rating THEN BEGIN
      { we're smaller }
      diff := their_rating - attack_rating + DIFF_OFFSET;
      divisor := DIVISOR_IF_SMALLER;
    END ELSE BEGIN
      { we're bigger then them. }
      diff := attack_rating - their_rating + DIFF_OFFSET;
      divisor := DIVISOR_IF_BIGGER;
    END;

    { Bonuses }
    Inc(divisor, eff_rating_bonus);
    IF i = f.enemy THEN Inc(divisor, DIVISOR_SAME_ENEMY_BONUS);
    IF their_ship^.status.damage > 0 THEN Inc(divisor, Trunc(their_ship^.status.damage));
    Inc(divisor, (100 - Trunc(their_ship^.status.shield)) DIV 5);

    IF f.data.speed <= 0 THEN multiplier := Trunc(DistanceFleetToShip(f, i) / 100)
                         ELSE multiplier := Trunc(DistanceFleetToShip(f, i) / f.data.speed); {?}
    IF multiplier < MULTIPLIER_MIN THEN multiplier := MULTIPLIER_MIN;

    diff := diff * multiplier DIV divisor;

    IF diff < best_diff THEN BEGIN
      best_diff := diff;
      best := i;
    END;
  END;

  f.enemy := best;
  IF f.enemy>=0 THEN BEGIN
    num_torpers := 0;
    FOR i:=1 TO f.data.num_ships DO
     WITH flak_ships^[f.data.first_ship + i-1].fs^ DO
      IF (status.is_alive) AND (data.torp_lcount > 0) AND (status.torp_count >= 10) THEN
       Inc(num_torpers);
    FOR i:=1 TO f.data.num_ships DO
     ComputeTorpLimit(flak_ships^[f.data.first_ship + i-1].fs^, flak_ships^[f.enemy].fs^, num_torpers);
  END;
END; { ChooseEnemy }

{ Launch fighters from fleet f }
PROCEDURE LaunchFighters(VAR f : FlakFleet);
VAR i, bay : INTEGER;
    sh     : FlakShipPtr;
    p      : FlakObjectPtr;
  pid      : INTEGER;
LABEL NextShip;
BEGIN
  IF (f.enemy<0) OR (NOT f.alive) OR (NOT flak_ships^[f.enemy].fs^.status.is_alive) THEN Exit;
  pid := f.data.player;

  FOR i := f.data.first_ship TO f.data.first_ship + f.data.num_ships-1 DO BEGIN
    sh := flak_ships^[i].fs;
    IF (sh^.status.is_alive) AND (sh^.data.bay_count>0) AND (sh^.status.fighter_count > 0)
     AND (sh^.status.LaunchCountdown=0)
     AND (sh^.status.fighters_launched < sh^.data.max_fighters_launched) THEN
    BEGIN
      FOR bay := 1 TO sh^.data.bay_count DO
       IF sh^.status.BayStatus[bay] >= 1000 THEN BEGIN
         { bay is ready }
         p := Malloc(Sizeof(FlakObject));
         IF p = NIL THEN BEGIN
           flak_oom := TRUE;
           Exit;
         END;
         p^.kind := oFighter;
         p^.can_change_enemy := TRUE;
         p^.x := f.x;
         p^.y := f.y;
         p^.z := sh^.status.z_offs;
         p^.enemy := f.enemy;
         p^.owner := i;
         p^.strikes := sh^.config.StrikesPerFighter;
         p^.kill := sh^.config.FighterBeamKill;
         p^.expl := sh^.config.FighterBeamExplosive;
         p^.death_flag := 1;
         p^.speed := sh^.config.FighterMovementSpeed;
         p^.next := NIL;

         sh^.status.BayStatus[bay] := 0;
         Dec(sh^.status.fighter_count);
         Inc(sh^.status.fighters_launched);
         sh^.status.LaunchCountdown := sh^.config.BayLaunchInterval;

         IF sh^.stat.min_ftr_here > sh^.status.fighter_count THEN
           sh^.stat.min_ftr_here := sh^.status.fighter_count;

         p^.next := players[pid].stuff;
         players[pid].stuff := p;

         GOTO NextShip; { only one fighter per tick }
       END;
    END;
   NextShip:;
  END;
END;

{ Fire torps from fleet f which belongs to player pid }
PROCEDURE FireTorps(VAR f : FlakFleet);
VAR
  sh, tl     : INTEGER;
  p          : FlakObjectPtr;
  psh        : FlakShipPtr;
  enef       : FlakFleetPtr;
  kill, dama : INTEGER;
  i          : INTEGER;
  pid        : INTEGER;
LABEL NextShip;
BEGIN
  IF (f.enemy < 0) OR NOT f.alive OR NOT (flak_ships^[f.enemy].fs^.status.is_alive) THEN Exit;

  pid := f.data.player;
  enef := FleetOf(f.enemy);

  { okay, torps can fire }
  FOR sh := f.data.first_ship TO f.data.first_ship + f.data.num_ships-1 DO BEGIN
    psh := flak_ships^[sh].fs;
    IF (psh^.status.is_alive) AND (psh^.status.torp_count > 0)
     AND IsDistanceLERadius(enef^.x, enef^.y, f.x, f.y, psh^.config.TorpFiringRange)
    THEN BEGIN
      FOR tl := 1 TO psh^.status.torp_limit DO
       IF psh^.status.TorpStatus[tl] >= 1000 THEN BEGIN
         { torpedo fires }
         p := Malloc(Sizeof(FlakObject));
         IF p = NIL THEN BEGIN
           flak_oom := TRUE;
           Exit;
         END;
         p^.kind := oTorpedo;
         p^.can_change_enemy := FALSE;
         p^.x := f.x;
         p^.y := f.y;
         p^.z := psh^.status.z_offs;
         p^.enemy := f.enemy;
         p^.owner := sh;
         p^.speed := TORP_MOVEMENT_SPEED;
         p^.next := NIL;
         p^.kill := Torps[psh^.data.torp_type].kill;
         p^.expl := Torps[psh^.data.torp_type].damage;
         p^.death_flag := Torps[psh^.data.torp_type].damage;
         IF psh^.config.AlternativeCombat=0 THEN BEGIN
           Inc(p^.kill, p^.kill);
           Inc(p^.expl, p^.expl);
         END;
         IF FlakRandom(100) < psh^.config.TorpHitOdds THEN BEGIN
           { torp hits its intended target }
           p^.strikes := 1;
         END ELSE BEGIN
           { misses. might be ricochet }
           p^.strikes := 0;
         END;
         psh^.status.TorpStatus[tl] := 0;
         Dec(psh^.status.torp_count);

         p^.next := players[pid].stuff;
         players[pid].stuff := p;

         GOTO NextShip; { only one torp per tick }
       END;
    END;
   NextShip:
  END;
END;

{ Fire fleet f's beams }
PROCEDURE FireBeams(VAR f : FlakFleet);
VAR
  sh, bm     : INTEGER;
  dist       : REAL;
  kill, dama : INTEGER;
  min_dist   : REAL;
  fdist      : REAL;
  min_ftr, p : FlakObjectPtr;
  pl         : INTEGER;
  psh        : FlakShipPtr;
  fl         : FlakFleetPtr;
  enef       : FlakFleetPtr;
LABEL NextShip;
BEGIN
  IF NOT f.alive THEN Exit;

  IF f.enemy>=0 THEN dist := DistanceFleetToShip(f, f.enemy)
                ELSE dist := 1.0e+15;

  { okay, beams can fire }
  FOR sh := f.data.first_ship TO f.data.first_ship + f.data.num_ships-1 DO BEGIN
    psh := flak_ships^[sh].fs;
    IF NOT psh^.status.is_alive THEN Continue;
    { first, try to fire at fighters }
    FOR bm := 1 TO psh^.data.beam_count DO BEGIN
      IF psh^.status.BeamStatus[bm] >= psh^.config.BeamHitFighterCharge THEN BEGIN
        { find a fighter }
        min_ftr := NIL;
        FOR pl := 1 TO FLAK_MAX_PLAYERS DO
          IF pl<>psh^.data.player THEN BEGIN
           p := players[pl].stuff;
           WHILE p<>NIL DO BEGIN
             IF (p^.kind = oFighter)
              AND (p^.enemy>=0) AND (p^.owner>=0) AND (f.enemy>=0)
              AND ((flak_ships^[p^.enemy].fs^.fleet_nr = psh^.fleet_nr)
                   OR (flak_ships^[p^.owner].fs^.fleet_nr = flak_ships^[f.enemy].fs^.fleet_nr))
             THEN BEGIN
               fdist := Sqrt(Sqr(f.x - p^.x + 0.0) + Sqr(f.y - p^.y + 0.0));
               { FIXME BeamHitFighterRange? }
               IF (fdist <= psh^.config.BeamFiringRange)
               { pick this fighter if
                 - we don't already have one
                 - we prefer firing on attacking fighters and only have a
                   retreating one so far
                 - it's closer than the one we already have }
                AND ((min_ftr=NIL)
                     OR ((psh^.config.FireOnAttackFighters>0)
                         AND (min_ftr^.strikes=0) AND (p^.strikes>0))
                     OR ((fdist < min_dist)
                         AND ((psh^.config.FireOnAttackFighters=0)
                              OR ((min_ftr^.strikes=0) AND (p^.strikes=0))
                              OR ((min_ftr^.strikes>0) AND (p^.strikes>0)))))
                THEN BEGIN
                  min_ftr := p;
                  min_dist := fdist;
                END;
             END;
             p := p^.next;
           END;
         END;
        { got a fighter. Fire. }
        IF min_ftr<>NIL THEN BEGIN
          VisDrawBeam(f.x, f.y, psh^.status.z_offs, min_ftr^.x, min_ftr^.y, min_ftr^.z);
          IF FlakRandom(100) < psh^.config.BeamHitOdds THEN BEGIN
            min_ftr^.kind := oDeleteMe;
            IF min_ftr^.owner>=0 THEN Dec(flak_ships^[min_ftr^.owner].fs^.status.fighters_launched);
          END;
          psh^.status.BeamStatus[bm] := 0;
          GOTO NextShip;
        END;
      END; { if beam charged }
    END; { for each beam }

    IF (f.enemy>=0) AND (flak_ships^[f.enemy].fs^.status.is_alive) AND (dist <= psh^.config.BeamFiringRange) THEN BEGIN
      FOR bm := 1 TO psh^.data.beam_count DO
       IF psh^.status.BeamStatus[bm] >= psh^.config.BeamHitShipCharge THEN BEGIN
         { beam fires }
         kill := Beams[psh^.data.beam_type].kill;
         dama := Beams[psh^.data.beam_type].damage;
         IF RaceIDs[psh^.data.player]=5 THEN kill := kill*3;
         IF FlakRandom(100) < psh^.config.BeamHitOdds THEN BEGIN
           HitShipWith(f.enemy, sh,
                       dama * LONGINT(psh^.status.BeamStatus[bm]) DIV 1000,
                       kill * LONGINT(psh^.status.BeamStatus[bm]) DIV 1000,
                       dama {death flag: 0 = death ray});
         END;
         psh^.status.BeamStatus[bm] := 0;
         fl := FleetOf(f.enemy);
         VisDrawBeam(f.x, f.y, psh^.status.z_offs,
                     fl^.x, fl^.y, flak_ships^[f.enemy].fs^.status.z_offs);
         GOTO NextShip;
       END;
    END;
   NextShip:
  END;
END;

{ Festivities are over when noone finds a matching opponent or
  everyone's dead. FIXME: this should be replaced by a more elaborate
  check, to detect things like "only freighters left" or "only planet
  and ship with death rays left". I guess this can be solved by
  formatting the attack lists appropriately. Think about it. }
FUNCTION FlakEndCheck:BOOLEAN;
VAR i,f : INTEGER;
BEGIN
  { optimisation idea: when there's only one player left, we don't
    need to wait for his fighters to return. }
  FOR i:=1 TO FLAK_MAX_PLAYERS DO BEGIN
    IF players[i].stuff<>NIL THEN BEGIN
      FlakEndCheck := FALSE;
      Exit;
    END;
  END;
  FOR f:=0 TO flak_fleet_cnt-1 DO BEGIN
    IF (flak_fleets^[f]^.alive) AND (flak_fleets^[f]^.enemy>=0) THEN BEGIN
      FlakEndCheck := FALSE;
      Exit;
    END;
  END;
  FlakEndCheck := TRUE;
END;

{ Compute new position for fleet f }
PROCEDURE ComputeNewPosition(VAR f : FlakFleet);
VAR pl     : INTEGER;
  sod      : LONGINT;
  x, y     : LONGINT;
  ps       : FlakShipPtr;
  dist, nd : REAL;
  living   : INTEGER;
  bvx, bvy : LONGINT;
  bvdivi   : REAL;
  pp, ff   : INTEGER;
  idist    : REAL;

CONST speed = 0.75;
BEGIN
  f.new_x := f.x;
  f.new_y := f.y;

  IF NOT f.alive THEN Exit; { fleet is dead }

  { if our enemy died, pick a new one }
  IF (f.enemy>=0) AND (NOT flak_ships^[f.enemy].fs^.status.is_alive) THEN
    ChooseEnemy(f);

  { now move 'em }
  { FIXME: not quite clean because it gets the standoff distance from the ship.
    But where else could we get that from? }
  sod := flak_ships^[f.data.first_ship].fs^.config.StandoffDistance;

  IF f.enemy >= 0 THEN BEGIN
    ps := flak_ships^[f.enemy].fs;
    x := flak_fleets^[ps^.fleet_nr]^.x;
    y := flak_fleets^[ps^.fleet_nr]^.y;

    dist := Sqrt(Sqr(0.0+(x-f.x))+Sqr(0.0+(y-f.y)));
    IF dist > 0 THEN BEGIN
      nd := dist - f.data.speed;
      IF nd < sod THEN nd := sod;
      IF nd < dist THEN BEGIN
        f.new_x := x + Round((f.x - x) * nd / dist);
        f.new_y := y + Round((f.y - y) * nd / dist);
      END;
    END;
  END;

  { compute bounce vector. }
  bvx := 0; bvy := 0; bvdivi := 0;
  pp := f.data.player;
  FOR ff:=0 TO flak_fleet_cnt-1 DO BEGIN
    IF (Addr(f) <> flak_fleets^[ff]) AND (flak_fleets^[ff]^.alive) THEN BEGIN
      idist := DistanceFleetToXY(f.new_x, f.new_y, flak_fleets^[ff]^);
      pl := flak_fleets^[ff]^.data.player;
      IF (idist>10) {arbitrary}
        AND (((pl=pp) AND (4*idist < sod))
             OR ((pl<>pp) AND (2*idist < sod)))
        THEN BEGIN
          Inc(bvx, f.new_x - flak_fleets^[ff]^.x);
          Inc(bvy, f.new_y - flak_fleets^[ff]^.y);
          bvdivi := bvdivi + idist;
        END;
    END;
  END;
  IF (bvx <> 0) OR (bvy <> 0) THEN BEGIN
    f.new_x := Round(f.new_x + bvx * speed * f.data.speed / bvdivi);
    f.new_y := Round(f.new_y + bvy * speed * f.data.speed / bvdivi);
  END;
END;

{ Garbage-collect fleet f. Kill all ships which are dead. Planet's
  weapon downgrades would happen here, too. }
PROCEDURE FleetGC(VAR f : FlakFleet);
VAR limit : INTEGER;
  living  : INTEGER;
  i       : INTEGER;
  player  : INTEGER;
  sh : FlakShipPtr;
BEGIN
  IF NOT f.alive THEN Exit;

  player := f.data.player;
  IF RaceIds[player]=2 THEN limit:=150 ELSE limit:=99;

  { kill all dead ships }
  living := 0;
  FOR i := f.data.first_ship TO f.data.first_ship+f.data.num_ships-1 DO BEGIN
    sh := flak_ships^[i].fs;
    IF (sh^.status.is_alive) THEN BEGIN
      IF (sh^.status.damage > limit) OR
        (NOT (sh^.data.flags AND flak_IsPlanet<>0) AND (sh^.status.crew < 0.5))
      THEN BEGIN
        {VIS:destroy ship}
        sh^.status.is_alive := FALSE;
        Dec(players[player].num_live_ships);
        Dec(players[player].sum_strength, sh^.data.compensation_rating);
      END ELSE BEGIN
        Inc(living);
      END;
    END;
  END;
  IF living=0 THEN f.alive := FALSE;
END;

{ Fighter intercept between two players }
PROCEDURE FighterIntercept(VAR a, b : FlakPlayer);
VAR pa, pb : FlakObjectPtr;

  FUNCTION TryIntercept(pa, pb : FlakObjectPtr) : BOOLEAN;
  VAR left_f, right_f, one_f, right_probab : INTEGER;
  BEGIN
    TryIntercept := FALSE;
    { first check whether we're close enough }
    IF (Abs(pa^.x - pb^.x) <= FIGHTER_INTERCEPT_RANGE)
     AND (Abs(pa^.y - pb^.y) <= FIGHTER_INTERCEPT_RANGE)
     AND (Sqr(pa^.x - pb^.x) + Sqr(pa^.y - pb^.y) <= LONGINT(FIGHTER_INTERCEPT_RANGE) * FIGHTER_INTERCEPT_RANGE)
    THEN BEGIN
      { okay, do we want it? Play a bit with probabilities:
          left_f  = 100x probability that "a" does not fire but "b" does
          right_r = ...
          one_f   = probability that exactly one fires
          right_probab = chance that it's b
        I'm unsure about this one; these are my reverse-engineered formulas
        from "pvcr.exe" version 3.2e :-) }
      left_f  := (100-flak_ships^[pa^.owner].fs^.config.FighterKillOdds) * flak_ships^[pb^.owner].fs^.config.FighterKillOdds;
      right_f := (100-flak_ships^[pb^.owner].fs^.config.FighterKillOdds) * flak_ships^[pa^.owner].fs^.config.FighterKillOdds;
      one_f := (left_f + right_f) DIV 100;
      right_probab := right_f DIV one_f;

      IF FlakRandom(100) < one_f THEN BEGIN
        { okay, one fires. }
        IF FlakRandom(100) < right_probab THEN BEGIN
          { delete a }
          VisDrawBeam(pb^.x, pb^.y, pb^.z, pa^.x, pa^.y, pa^.z);
          pa^.kind := oDeleteMe;
          IF pa^.owner>=0 THEN Dec(flak_ships^[pa^.owner].fs^.status.fighters_launched);
        END ELSE BEGIN
          { delete b }
          VisDrawBeam(pa^.x, pa^.y, pa^.z, pb^.x, pb^.y, pb^.z);
          pb^.kind := oDeleteMe;
          IF pb^.owner>=0 THEN Dec(flak_ships^[pb^.owner].fs^.status.fighters_launched);
        END;
        TryIntercept := TRUE;
      END;
    END;
  END;

LABEL NextA;
BEGIN
  pa := a.stuff;
  { PVCR has the unfairness that when a "right" fighter kills a "left"
    one, it is permitted to fire again. This algorithm permits that
    goof for both sides. Why not. There is no way to make this 100%
    fair.
    -- no longer true, we have at most one FI per tick. }
  WHILE pa<>NIL DO BEGIN
    IF pa^.kind = oFighter THEN BEGIN
      pb := b.stuff;
      WHILE pb<>NIL DO BEGIN
        IF pb^.kind = oFighter THEN BEGIN
          { two fighters. Possible targets? }
          IF (pa^.owner>=0) AND (pb^.owner>=0) THEN
           IF (pa^.enemy = pb^.owner) OR (pb^.enemy = pa^.owner) THEN BEGIN
             IF TryIntercept(pa, pb) THEN Exit; { only one FI per tick }
             IF pa^.kind <> oFighter THEN GOTO NextA;
           END;
        END;
        pb := pb^.Next;
      END;
    END;
   NextA:
    pa := pa^.Next;
  END;
END;

{ All fighters fire at their enemies }
PROCEDURE FightersFire(VAR pl : FlakPlayer);
VAR
  p  : FlakObjectPtr;
  fl : FlakFleetPtr;
BEGIN
  p := pl.stuff;
  WHILE p<>NIL DO BEGIN
    IF (p^.kind=oFighter) AND (p^.strikes>0) AND (flak_ships^[p^.enemy].fs^.status.is_alive) THEN BEGIN
      fl := flak_fleets^[flak_ships^[p^.enemy].fs^.fleet_nr];
      IF IsDistanceLERadius(fl^.x, fl^.y, p^.x, p^.y,
                            flak_ships^[p^.owner].fs^.config.FighterFiringRange)
      THEN BEGIN
        HitShipWith(p^.enemy, p^.owner, p^.kill, p^.expl, p^.death_flag);
        VisDrawBeam(p^.x, p^.y, p^.z, fl^.x, fl^.y, flak_ships^[p^.enemy].fs^.status.z_offs);
        Dec(p^.strikes);
        p^.can_change_enemy := FALSE;
      END;
    END;
    p := p^.next;
  END;
END;

{ Find a new base for fighter /p/ }
PROCEDURE FindNewBase(VAR pl : FlakPlayer; p : FlakObjectPtr);
VAR
  i, j, id      : INTEGER;
  dist, mindist : REAL;
  minid         : INTEGER;
  minmass       : INTEGER;
  fl            : FlakFleetPtr;
BEGIN
  mindist := 1.0E+15;
  minid   := -1;
  { choose closest fleet which has a carrier }
  FOR i:=0 TO flak_fleet_cnt-1 DO BEGIN
    fl := flak_fleets^[i];
    IF fl^.data.player <> pl.number THEN Continue;
    dist := Sqrt(Sqr(fl^.x - p^.x + 0.0) + Sqr(fl^.y - p^.y + 0.0));
    IF dist < mindist THEN BEGIN
      minmass := 0;
      FOR j:=1 TO fl^.data.num_ships DO BEGIN
        id := j + fl^.data.first_ship-1;
        IF (flak_ships^[id].fs^.status.is_alive) AND (flak_ships^[id].fs^.data.bay_count > 0)
          AND (flak_ships^[id].fs^.data.mass > minmass) THEN BEGIN
            minmass := flak_ships^[id].fs^.data.mass;
            mindist := dist;
            minid   := id;
          END;
      END;
    END;
  END;
  IF minid>=0 THEN Inc(flak_ships^[minid].fs^.status.fighters_launched);
  p^.owner := minid;
END;

{ Move space stuff (torps/fighters) for player pl }
PROCEDURE MoveStuff(VAR pl : FlakPlayer);
VAR p : FlakObjectPtr;
  l   : LONGINT;
  f   : FlakFleetPtr;
  ene : INTEGER;
BEGIN
  p := pl.stuff;
  WHILE p<>NIL DO BEGIN
    IF p^.kind = oTorpedo THEN BEGIN
      f := FleetOf(p^.enemy);       { no matter whether it's dead }
      l := FlakObjectMoveTowards(p^, f^.x, f^.y, flak_ships^[p^.enemy].fs^.status.z_offs, p^.speed);
      IF l = 0 THEN BEGIN
        { torp hits }
        p^.kind := oDeleteMe;
        IF p^.strikes<>0 THEN BEGIN
          ene := p^.enemy;
          IF (NOT flak_ships^[ene].fs^.status.is_alive) AND (f^.alive) THEN
            ene := f^.data.first_ship + FlakRandom(f^.data.num_ships);
          IF flak_ships^[ene].fs^.status.is_alive THEN
            HitShipWith(ene, p^.owner, p^.expl, p^.kill, p^.death_flag);
          IF p^.owner >= 0 THEN
            Inc(flak_ships^[p^.owner].fs^.stat.torps_hit);
        END;
      END;
    END ELSE IF p^.kind = oFighter THEN BEGIN
      { fighter }
      { NOTE: do *not* choose a new base unconditionally. If that is
        done, we must change the fighter firing code, too, which
        refers to the base
        NOTE 07/Nov/04: huh? }
      IF (p^.strikes > 0) AND NOT flak_ships^[p^.enemy].fs^.status.is_alive THEN BEGIN
        { okay, enemy is dead }
        IF NOT p^.can_change_enemy THEN BEGIN
          { we cannot change our enemy, retreat. }
          p^.strikes := 0;
        END ELSE IF (p^.owner>=0) AND (flak_ships^[p^.owner].fs^.status.is_alive) THEN BEGIN
          { our base is still alive, ask it what to do. }
          IF (FleetOf(p^.owner)^.enemy <> p^.enemy) THEN BEGIN
            p^.enemy := FleetOf(p^.owner)^.enemy;
          END ELSE BEGIN
            { base has not decided yet. Wait. }
            Dec(p^.strikes);
          END;
        END ELSE BEGIN
          { base is dead. Pick a new one, and go for its enemy. This
            might also be a dead one (because death is declared after
            this part of the battle sequence), but we don't care. }
          FindNewBase(pl, p);
          IF p^.owner>=0 THEN p^.enemy := FleetOf(p^.owner)^.enemy
                         ELSE p^.enemy := -1;
        END;
      END;

      IF (p^.enemy<0) OR (pl.num_live_ships = 0) THEN BEGIN
        p^.strikes := 0;
      END;

      IF p^.strikes = 0 THEN BEGIN
        { return to base }
        IF (p^.owner<0) OR (NOT flak_ships^[p^.owner].fs^.status.is_alive) THEN
          FindNewBase(pl, p);
        IF (p^.owner<0) THEN BEGIN
          p^.kind := oDeleteMe;
        END ELSE BEGIN
          f := FleetOf(p^.owner);
          IF FlakObjectMoveTowards(p^, f^.x, f^.y, flak_ships^[p^.owner].fs^.status.z_offs, p^.speed)=0 THEN BEGIN
            { fighter reached baseship }
            p^.kind := oDeleteMe;
            Inc(flak_ships^[p^.owner].fs^.status.fighter_count);
            Dec(flak_ships^[p^.owner].fs^.status.fighters_launched);
          END;
        END;
      END ELSE BEGIN
        { move towards enemy }
        f := FleetOf(p^.enemy);
        FlakObjectMoveTowards(p^, f^.x, f^.y, flak_ships^[p^.enemy].fs^.status.z_offs, p^.speed);
      END;
    END;
    p := p^.next;
  END;
END;

{ Initialize everything. If battle was already advanced a bit,
  move back to beginning.
  NOTE: this does NOT free objects occupied by FlakObjects; use
  FlakBattleDone first. }
PROCEDURE FlakBattleInit;
VAR i, sh, n : INTEGER;
  cur_z      : INTEGER;
  fl         : FlakFleetPtr;
BEGIN
  flak_time := 0;
  flak_rng_seed := flak_header_seed;

  FOR i := 1 TO FLAK_MAX_PLAYERS DO
   FlakPlayerInit(i);
  FOR i := 0 TO flak_fleet_cnt-1 DO BEGIN
    fl := flak_fleets^[i];
    cur_z := (fl^.data.num_ships-1) * (-50);
    FOR sh := 1 TO fl^.data.num_ships DO BEGIN
      n := sh + fl^.data.first_ship - 1;
      FlakShipInit(flak_ships^[n].fs^, n, i);
      flak_ships^[n].fs^.status.z_offs := cur_z;
      Inc(cur_z, 100);
      Inc(players[fl^.data.player].num_live_ships);
      Inc(players[fl^.data.player].sum_strength,
          flak_ships^[n].fs^.data.compensation_rating);
    END;
    FlakFleetInit(fl^);
    fl^.alive := TRUE;
  END;

  FOR i := 0 TO flak_fleet_cnt-1 DO
    ChooseEnemy(flak_fleets^[i]^);
END; { FlakBattleInit }

{ Shut down player }
PROCEDURE FlakBattleDone;
VAR i : INTEGER;
BEGIN
  FOR i := 1 TO FLAK_MAX_PLAYERS DO
    FlakPlayerDone(i);
END; { FlakBattleDone }

{ Print a checkpoint. This is used for debugging. The C++ implementation
  contains a similar function which outputs identical values. }
PROCEDURE FlakPrintCheckpoint;

  PROCEDURE WriteUnit(i : INTEGER);
  BEGIN
    IF i < 0 THEN Write('none') ELSE Write('#', flak_ships^[i].fs^.data.id);
  END;

VAR
  i, j : INTEGER;
  p    : FlakObjectPtr;
BEGIN
  Writeln('=== ', flak_time, ' (seed=', flak_rng_seed, ') ===');
  FOR i := 0 TO flak_fleet_cnt-1 DO
   WITH flak_fleets^[i]^ DO BEGIN
     Write('fleet ', i, ': pos=', x, ',', y, ' alive=', alive,
             ' enemy=');
     WriteUnit(enemy);
     Writeln;
   END;
  FOR i := 0 TO flak_ship_cnt-1 DO
   WITH flak_ships^[i].fs^ DO BEGIN
     Write('ship ', i, ': crew=', Trunc(status.crew),
           ' damage=', Trunc(status.damage),
           ' shield=', Trunc(status.shield),
           ' alive=', status.is_alive,
           ' bs=');
     FOR j:=1 TO 5 DO Write(status.BeamStatus[j], ' ');
     Write('ts=');
     FOR j:=1 TO 5 DO Write(status.TorpStatus[j], ' ');
     Write('fs=');
     FOR j:=1 TO 5 DO Write(status.BayStatus[j], ' ');
     Writeln('ftrs_out=', status.fighters_launched,
             ' ftrs_in=', status.fighter_count,
             ' cdn=', status.launchcountdown,
             ' torps=', status.torp_count);
   END;
  FOR i := 1 TO 11 DO BEGIN
    Writeln('player ', i, ': num_live_ships=',
            players[i].num_live_ships);
    p := players[i].stuff;
    WHILE p<>NIL DO BEGIN
      Write('  ', INTEGER(p^.kind), ' at ',
            p^.x, ',', p^.y, ' ene=');
      WriteUnit(p^.enemy);
      Write(' owner=');
      WriteUnit(p^.owner);
      Writeln(' kill=', p^.kill, ' expl=', p^.expl,
              ' speed=', p^.speed, ' strikes=', p^.strikes,
              ' can_change=', p^.can_change_enemy);
      p := p^.next;
    END;
  END;
END;

{ Main routine: play one battle tick }
FUNCTION FlakPlayOneTick : BOOLEAN;
VAR i,j,f,fleet_off,player_off:INTEGER;
BEGIN
  FlakPlayOneTick := FALSE;
  FOR i:=1 TO flak_ship_cnt DO
   FlakShipRecharge(flak_ships^[i-1].fs^);
  IF (flak_time<>0) AND ((flak_time MOD CHOOSE_ENEMY_TIME) = 0) THEN
   FOR i:=1 TO flak_fleet_cnt DO
    ChooseEnemy(flak_fleets^[i-1]^);
  FOR i:=1 TO flak_fleet_cnt DO
   LaunchFighters(flak_fleets^[i-1]^);
  FOR i:=1 TO flak_fleet_cnt DO
   FireTorps(flak_fleets^[i-1]^);

  fleet_off := FlakRandom(flak_fleet_cnt);
  FOR i:=fleet_off TO flak_fleet_cnt-1 DO
    FireBeams(flak_fleets^[i]^);
  FOR i:=0 TO fleet_off-1 DO
    FireBeams(flak_fleets^[i]^);

  player_off := FlakRandom(FLAK_MAX_PLAYERS);
  FOR i:=player_off+1 TO FLAK_MAX_PLAYERS DO
   FightersFire(players[i]);
  FOR i:=1 TO player_off DO
   FightersFire(players[i]);

  FOR i:=1 TO FLAK_MAX_PLAYERS-1 DO
   FOR j:=i+1 TO FLAK_MAX_PLAYERS DO
    FighterIntercept(players[i], players[j]);

  FOR i:=player_off+1 TO FLAK_MAX_PLAYERS DO
   MoveStuff(players[i]);
  FOR i:=1 TO player_off DO
   MoveStuff(players[i]);

  FOR i:=1 TO flak_fleet_cnt DO
   FleetGC(flak_fleets^[i-1]^);
  FOR i:=1 TO FLAK_MAX_PLAYERS DO
   FlakPlayerGC(players[i]);
  FOR i:=1 TO flak_fleet_cnt DO
    ComputeNewPosition(flak_fleets^[i-1]^);
  FOR i:=1 TO flak_fleet_cnt DO
    WITH flak_fleets^[i-1]^ DO BEGIN x := new_x; y := new_y; END;

  {FlakPrintCheckpoint;}

  Inc(flak_time);

  FlakPlayOneTick := FlakEndCheck;
END;

{=============================== VISUALISATION ==========================}

TYPE FlakVisMode = (vis_Flat, vis_Iso, vis_3D);

TYPE
  FlakScreenPos = RECORD
                    x, y : INTEGER;
                  END;
  FlakListPos   = RECORD
                    fleet    : INTEGER;     { fleet number }
                    ship_rel : INTEGER;     { relative ship number; -1 if it identifies the fleet itself }
                  END;

TYPE FlakVisStatus = (vs_Halt, vs_Playing, vs_Terminated, vs_Revived);

CONST
  vis_mode    : FlakVisMode = vis_Flat;
  vis_grid    : BOOLEAN     = TRUE;
  vis_panel   : BOOLEAN     = FALSE;

VAR
  vis_max_r     : LONGINT;
  vis_grid_size : LONGINT;
  vis_status    : FlakVisStatus;
  vis_show      : ARRAY[0..1] OF INTEGER;

{ 3D parameters }
CONST
  t_dist            = 50000;                { distance from screen to viewer }

  t_cam_azi : REAL  = 0;                    { camera azimut }
  t_cam_hei : REAL  = 0;                    { camera height }

  t_org_x : LONGINT = 0;                    { camera is watching this point }
  t_org_y : LONGINT = 0;
  t_org_z : INTEGER = 0;

  t_scale           = 240/20000;
  t_z_scale         = 2000/100;

CONST
  PANEL_HEIGHT      = 100;

VAR
  t_follow_fleet : INTEGER;

{ Compute new value for vis_max_r }
FUNCTION VisComputeRange : LONGINT;
VAR
  i : INTEGER;
  l : LONGINT;
BEGIN
  l := 10000;
  FOR i := 1 TO flak_fleet_cnt DO
    WITH flak_fleets^[i-1]^ DO BEGIN
      IF Abs(x) > l THEN l := Abs(x);
      IF Abs(y) > l THEN l := Abs(y);
    END;
  VisComputeRange := l + 2000;
END;

{ Initialize visualizer coordinates }
PROCEDURE VisInit;
VAR i : INTEGER;
BEGIN
  vis_max_r := VisComputeRange;
  vis_grid_size := vis_max_r;
  vis_status := vs_Halt;

  vis_panel := FALSE;
  vis_show[0] := 0;
  vis_show[1] := 0;

  t_cam_azi := 0;
  t_cam_hei := 0;

  t_org_x := 0;
  t_org_y := 0;
  t_org_z := 0;

  t_follow_fleet := -1;
  FOR i := 0 TO flak_fleet_cnt-1 DO
    IF flak_fleets^[i]^.data.player = Global.PlayerID THEN BEGIN
      t_follow_fleet := i;
      Break;
    END;
END; { VisInit }

{ Convert coordinates to screen format }
FUNCTION VisConvertCoords(x,y:LONGINT; z:INTEGER; VAR sp:FlakScreenPos) : BOOLEAN;
VAR vp, xp, yp, zp : REAL;
    cosa,sina      : REAL;
    cosg,sing      : REAL;
    yorg           : INTEGER;

  FUNCTION Wrap(r:REAL):INTEGER;
  BEGIN
    IF (r < -2500) OR (r > 2500) THEN Wrap:=-10000 ELSE
     Wrap := Round(r);
  END;

BEGIN
  IF NOT vis_panel THEN yorg := 240 ELSE yorg := 240 - PANEL_HEIGHT DIV 2;
  CASE vis_mode OF
    vis_Flat : BEGIN
      sp.x := 240  + Round(x * 240 / vis_max_r);
      sp.y := yorg - Round(y * 240 / vis_max_r);
      VisConvertCoords := TRUE;
    END;
    vis_Iso  : BEGIN
      sp.x := 240        + Round(150 * (x - y) / vis_max_r);
      sp.y := yorg + 80 -       150 * (x + y) DIV (3 * vis_max_r) + (LONGINT(1500) * z DIV vis_max_r);
      VisConvertCoords := TRUE;
    END;
    vis_3D   : BEGIN
      vp := Sqrt(Sqr(0.0+x) + Sqr(0.0+y));

      cosa := Cos(t_cam_azi);
      sina := Sin(t_cam_azi);
      cosg := Cos(t_cam_hei);
      sing := Sin(t_cam_hei);

      IF vp>0 THEN BEGIN
        xp := (x*cosa - y*sina);
        yp := (x*sina + y*cosa);
      END ELSE BEGIN
        xp := 0;
        yp := 0;
      END;
      zp := (yp*cosg + 1.0*z*t_z_scale*sing);
      yp := (yp*sing - 1.0*z*t_z_scale*cosg);

      xp := xp - t_org_x;
      yp := yp - t_org_y;
      zp := zp - t_org_z;

      { Now we have the y axis heading straight away, and
        the x/z plane in front of us. }

      { t_dist / h' = (t_dist + t_cam_dist - yp) / x }
      sp.x := Wrap(240  + t_scale * xp * t_dist / (t_dist + vis_max_r - yp));
      sp.y := Wrap(yorg - t_scale * zp * t_dist / (t_dist + vis_max_r - yp));
      VisConvertCoords := (sp.x <> -10000) AND (sp.y <> -10000)
    END;
  END;
END;

{ Draw a beam }
PROCEDURE VisDrawBeam(x1,y1:LONGINT; z1:INTEGER; x2,y2:LONGINT; z2:INTEGER);
VAR b, e : FlakScreenPos;
BEGIN
  IF vis_inhibit THEN Exit;
  Color := Red;
  IF VisConvertCoords(x1,y1,z1,b) AND VisConvertCoords(x2,y2,z2,e) THEN
   Line(b.x,b.y,e.x,e.y);
END; { VisDrawBeam }

PROCEDURE VisDrawCharge(x, y, wi, ch : INTEGER);
VAR
  c  : BYTE;
BEGIN
  {  x + 112, y + 11 + 2*i, 116, ps^.status.BeamStatus[i];}
  ch := ch DIV 8; { FIXME? 8 * 116 = 928 }
  IF ch > wi THEN ch := wi;

  IF _col256 THEN BEGIN
    c := 112 + ch SHR 2;
    IF c>127 THEN c:=127;
    Color:=(c);
    HLine(x, y, x + ch);
  END ELSE BEGIN
    CASE ch OF
      0..10:c:=Red;
      11..20:c:=Red+16*Yellow;
      21..30:c:=Yellow;
      31..50:c:=Yellow+16*Green;
    ELSE c:=Green;
    END;
    ExtdBar(x, y, x + ch, y, c);
  END;
END;

PROCEDURE VisDrawShipPanel(x, y, sh, width : INTEGER);
VAR
  ps                      : FlakShipPtr;
  xw, panel_height, yw, i : INTEGER;

  FUNCTION AdjustPanelPosition : BOOLEAN;
  BEGIN
    IF yw + panel_height > 475 THEN BEGIN
      yw := y + 2;
      Inc(xw, 122);
    END;
    AdjustPanelPosition := (xw < x + width - 120);
  END;

BEGIN
  ps := flak_ships^[sh].fs;
  TextStyle := 3;
  JustX := 0;
  JustY := 0;
  OutLTextXY(x+2, y+2, width-4, GetStr(ps^.data.name, 20));
  TextStyle := 2;
  IF ps^.data.flags AND flak_IsPlanet<>0 THEN
    OutTextXY(x+2, y+17, 'Planet')
  ELSE
    OutTextXY(x+2, y+17, GetStr(Hulls^[ps^.data.hull].Name, 30));
  OutTextXY(x+2, y+32, 'Damage:');
  OutTextXY(x+2, y+44, 'Shield:');
  OutTextXY(x+2, y+56, 'Crew:');   
  JustX := 2;
  OutTextXY(x+100, y+32, itoa(Round(ps^.status.damage)) + '%');
  OutTextXY(x+100, y+44, itoa(Round(ps^.status.shield)) + '%');
  IF ps^.data.flags AND flak_IsPlanet=0 THEN
    OutTextXY(x+100, y+56, itoa(Round(ps^.status.crew)));
  JustX := 0;

  IF vis_status = vs_Revived THEN BEGIN
    IF ps^.data.ending_status < 0 THEN BEGIN
      OutTextXY(x+2, y+70, 'Destroyed');
    END ELSE IF ps^.data.ending_status = 0 THEN BEGIN
      OutTextXY(x+2, y+70, 'Survived');
    END ELSE IF ps^.data.ending_status = ps^.data.player THEN BEGIN
      OutTextXY(x+2, y+70, 'Captured back');
    END ELSE BEGIN
      OutTextXY(x+2, y+70, 'Captured by ' + itoa(ps^.data.player));
    END;
  END ELSE IF ps^.status.is_alive THEN BEGIN
    OutTextXY(x+2, y+70, 'Alive');
  END ELSE BEGIN
    OutTextXY(x+2, y+70, 'Dead');
  END;

  { weapons }
  yw := 999;
  xw := x + 113;
  IF ps^.data.beam_count<>0 THEN BEGIN
    panel_height := 12 + 2 * ps^.data.beam_count;
    IF AdjustPanelPosition THEN BEGIN
      FrameDown (xw,     yw,   xw + 120, yw + panel_height);
      ExtdBar   (xw + 1, yw+1, xw + 119, yw + panel_height-1, 0);
      Color := White;
      OutLTextXY(xw + 2, yw+1, 117, GetStr(Beams[ps^.data.beam_type].Name, 20));
      FOR i := 1 TO ps^.data.beam_count DO BEGIN
        VisDrawCharge(xw + 1, yw + 10 + 2*i, 117, ps^.status.BeamStatus[i]);
      END;
      Inc(yw, panel_height + 2);
    END;
  END;

  IF ps^.data.torp_lcount<>0 THEN BEGIN
    panel_height := 24 + 2 * ps^.data.torp_lcount;
    IF AdjustPanelPosition THEN BEGIN
      FrameDown (xw,     yw,   xw + 120, yw + panel_height);
      ExtdBar   (xw + 1, yw+1, xw + 119, yw + panel_height-1, 0);
      Color := White;
      OutLTextXY(xw + 2, yw+1, 117, GetStr(Torps[ps^.data.torp_type].Name, 20));
      FOR i := 1 TO ps^.data.torp_lcount DO BEGIN
        VisDrawCharge(xw + 1, yw + 10 + 2*i, 117, ps^.status.TorpStatus[i]);
      END;
      IF ps^.status.torp_count > 5 THEN Color := White ELSE
      IF ps^.status.torp_count > 0 THEN Color := Gray ELSE Color := Dark;
      OutTextXY(xw + 2, yw + panel_height - 11, 'Torpedoes: ' + itoa(ps^.status.torp_count));
      Inc(yw, panel_height + 2);
    END;
  END;

  IF ps^.data.bay_count<>0 THEN BEGIN
    panel_height := 24;
    IF AdjustPanelPosition THEN BEGIN
      FrameDown (xw,     yw,   xw + 120, yw + panel_height);
      ExtdBar   (xw + 1, yw+1, xw + 119, yw + panel_height-1, 0);
      Color := White;
      IF ps^.data.bay_count<>1 THEN
        OutLTextXY(xw+2, yw+1, 117, itoa(ps^.data.bay_count) + ' fighter bays')
      ELSE
        OutLTextXY(xw+2, yw+1, 117, '1 fighter bay');
      IF ps^.status.fighter_count > 5 THEN Color := White ELSE
      IF ps^.status.fighter_count > 0 THEN Color := Gray ELSE Color := Dark;
      OutTextXY(xw + 2, yw + panel_height - 11, 'Fighters: ' + itoa(ps^.status.fighter_count));
      Inc(yw, panel_height + 2);
    END;
  END;
END;

{ Draw whole battlefield }
PROCEDURE VisDrawBattlefield;
VAR
  x, y, n  : INTEGER;
  a, b     : FlakScreenPos;
  pt       : FlakObjectPtr;
  i, f, sh : INTEGER;
  s1       : STRING[3];
  s2, sp   : STRING[7]; { ' (100%)' }
  fun      : PROCEDURE (x,y,r:INTEGER);
  rad      : INTEGER;
CONST
  PANEL_Y0 = 480 - PANEL_HEIGHT;
BEGIN
  { grid }
  IF vis_panel THEN Dec(WinY2, PANEL_HEIGHT);
  IF vis_grid THEN BEGIN
    SetColor(Dark);
    SetLineStyle($55);
    n := (vis_grid_size DIV 10000) + 1;
    FOR x:=-n TO n DO BEGIN
      IF VisConvertCoords(LONGINT(10000)*x, LONGINT(-10000)*n, 0, a)
       AND VisConvertCoords(LONGINT(10000)*x, LONGINT(+10000)*n, 0, b)
        THEN Line(a.x, a.y, b.x, b.y);
      IF VisConvertCoords(LONGINT(-10000)*n, LONGINT(10000)*x, 0, a)
       AND VisConvertCoords(LONGINT(+10000)*n, LONGINT(10000)*x, 0, b)
        THEN Line(a.x, a.y, b.x, b.y);
    END;
    SetLineStyle($FF);
  END;

  { battlefield }
  SetFillStyle(SOLIDFILL);
  FOR f := 0 TO flak_fleet_cnt-1 DO BEGIN
    Color := 15;
    IF NOT flak_fleets^[f]^.alive THEN Color := Dark;
    IF f = t_follow_fleet THEN BEGIN
      IF _col256 AND (Color = Dark) THEN Color := $8B ELSE Color := Yellow;
    END;
    IF flak_ships^[flak_fleets^[f]^.data.first_ship].fs^.data.flags AND flak_IsPlanet<>0 THEN BEGIN
      fun := FillCircle;
      rad := 7;
    END ELSE BEGIN
      fun := Circle;
      rad := 4;
    END;
    IF vis_mode = vis_Flat THEN BEGIN
      IF VisConvertCoords(flak_fleets^[f]^.x, flak_fleets^[f]^.y, 0, a) THEN
        fun(a.x, a.y, rad);
    END ELSE BEGIN
      FOR i:=1 TO flak_fleets^[f]^.data.num_ships DO BEGIN
        sh := flak_fleets^[f]^.data.first_ship + i - 1;
        IF flak_ships^[sh].fs^.status.is_alive THEN BEGIN
          IF VisConvertCoords(flak_fleets^[f]^.x, flak_fleets^[f]^.y,
                              flak_ships^[sh].fs^.status.z_offs, a)
          THEN fun(a.x, a.y, rad);
        END;
      END;
    END;
  END;

  FOR i:=1 TO FLAK_MAX_PLAYERS DO BEGIN
    pt := players[i].stuff;
    WHILE pt<>NIL DO BEGIN
      CASE pt^.kind OF
        oFighter: BEGIN
          IF (pt^.owner>=0) AND (flak_ships^[pt^.owner].fs^.status.is_alive)
           THEN Color := BlueGray
           ELSE Color := BlueBlack;
          IF VisConvertCoords(pt^.x, pt^.y, pt^.z, a) THEN
           Circle(a.x, a.y, 2);
        END;
        oTorpedo: BEGIN
          Color := Red;
          IF VisConvertCoords(pt^.x, pt^.y, pt^.z, a) THEN BEGIN
            HLine(a.x-3, a.y, a.x+3);
            VLine(a.x, a.y-3, a.y+3);
          END;
        END;
      END;
      pt := pt^.next;
    END;
  END;

  Str(flak_time DIV 60, s2);
  Str(flak_time MOD 60 : 2, s1);
  IF s1[1] = ' ' THEN s1[1] := '0';
  sp := '';
  IF flak_header_time > 0 THEN BEGIN
    Str(flak_time * 100 DIV flak_header_time, sp);
    sp := ' (' + sp + '%)';
  END;
  JustX := 0;
  JustY := 0;
  TextStyle := 1;
  Color := White;
  IF vis_panel THEN y := 480 - 16 - PANEL_HEIGHT ELSE y := 480 - 16;
  OutTextXY(0, y, 'Time: ' + s2 + ':' + s1 + sp);

  IF vis_panel THEN BEGIN
    Inc(WinY2, PANEL_HEIGHT);
    IF vis_show[0] = vis_show[1] THEN BEGIN
      { one panel }
      Frame(0, PANEL_Y0, 479, 479);
      ExtdBar(1, PANEL_Y0+1, 478, 478, Gray);
      VisDrawShipPanel(0, PANEL_Y0, vis_show[0], 480);
    END ELSE BEGIN
      { two panels }
      FOR i:=0 TO 1 DO BEGIN
        x := i * 240;
        Frame(x, PANEL_Y0, x + 239, 479);
        ExtdBar(x+1, PANEL_Y0+1, x + 238, 478, Gray);
        VisDrawShipPanel(x, PANEL_Y0, vis_show[i], 240);
      END;
    END;
  END;
  TextStyle := 0;
END; { VisDrawBattlefield }

{ Draw fixed parts of status panel at the given Y }
PROCEDURE VisDrawStatusPanel(y : INTEGER);
CONST x_org = 480 + 5 + 16*2 + 5;
      x_width = 638 - x_org;
BEGIN
  Frame(480, y, 639, y + 479);
  ExtdBar(481, y+1, 638, y + 478, Gray);
    { format:
      [p]    play/pause
      [+][-] speed
      [i][3] display
      [c]    cam
      [a][z] zoom
      [esc]  exit }
  TextStyle := 3;
  JustX := 0;
  JustY := 0;
  Color := Black;
  OutFTextXY(x_org, y + (475-1*16), x_width, 'Exit');
  OutFTextXY(x_org, y + (475-2*16), x_width, 'Manual Zoom');
  OutFTextXY(x_org, y + (475-3*16), x_width, 'Auto Camera');
  OutFTextXY(x_org, y + (475-4*16), x_width, 'Display Mode');
  OutFTextXY(x_org, y + (475-5*16), x_width, 'Speed');
  OutFTextXY(x_org, y + (475-6*16), x_width, 'Play/Pause');
  OutFTextXY(480 + 5, y + 5, 638-480-5, 'FLAK: ' + itoa(flak_ship_cnt) + ' units');
  HLine(480 + 5, y + (5 + 13), 635);
  TextStyle := 0;
END; { VisDrawStatusPanel }

{ Camera following in full 3D mode }
PROCEDURE CamFollow;

  FUNCTION WrapAngle(r : REAL):REAL;
  BEGIN
    IF r < -PI THEN WrapAngle := r + 2*PI ELSE
    IF r >  PI THEN WrapAngle := r - 2*PI ELSE
      WrapAngle := r;
  END;

VAR
  target_azi : REAL;
  target_he  : REAL;
  a, b, pl   : INTEGER;
BEGIN
  target_azi := 20 * PI / 180;
  target_he := -60 * PI / 180;
  IF (t_follow_fleet >= 0) AND NOT (flak_fleets^[t_follow_fleet]^.alive) THEN BEGIN
    b := -1;
    pl := flak_fleets^[t_follow_fleet]^.data.player;
    FOR a := 1 TO flak_fleet_cnt-1 DO
      IF flak_fleets^[a]^.alive THEN
        IF (b < 0)
          OR ((flak_fleets^[a]^.data.player = pl)
              AND (flak_fleets^[b]^.data.player <> pl))
          THEN b := a;
    IF b >= 0 THEN t_follow_fleet := b;
  END;
  IF (t_follow_fleet >= 0) AND (flak_fleets^[t_follow_fleet]^.alive) THEN BEGIN
    WITH flak_fleets^[t_follow_fleet]^ DO BEGIN
      IF (x<>0) OR (y<>0) THEN
        target_azi := ArcTan2(-x, -y) * (PI/180) + 20 * PI / 180
      ELSE
        Exit;
    END;
  END;
  target_azi := WrapAngle(target_azi - t_cam_azi);
  target_he := WrapAngle(target_he - t_cam_hei);
  a := Trunc(Abs(target_azi) * 33.0);
  b := Trunc(Abs(target_he) * 33.0);
  IF a > b THEN b := a;
  IF b < 10 THEN b := 10;
  t_cam_azi := t_cam_azi + target_azi / b;
  t_cam_hei := t_cam_hei + target_he / b;
END; { CamFollow }

{ Increment a list position by /delta/. Sticky at end, i.e. never
  overflows. }
PROCEDURE FlakListInc(VAR p : FlakListPos; delta : INTEGER);
BEGIN
  Inc(p.ship_rel, delta);
  WHILE p.ship_rel < -1 DO BEGIN
    IF p.fleet = 0 THEN BEGIN
      p.ship_rel := -1;
      Exit;
    END;
    Dec(p.fleet);
    Inc(p.ship_rel, flak_fleets^[p.fleet]^.data.num_ships + 1);
  END;
  WHILE p.ship_rel >= flak_fleets^[p.fleet]^.data.num_ships DO BEGIN
    IF p.fleet = flak_fleet_cnt-1 THEN BEGIN
      p.ship_rel := flak_fleets^[p.fleet]^.data.num_ships-1;
      Exit;
    END;
    Dec(p.ship_rel, flak_fleets^[p.fleet]^.data.num_ships + 1);
    Inc(p.fleet);
  END;
END; { FlakListInc }

{ Compare two list positions. }
FUNCTION FlakListLess(a, b : FlakListPos) : BOOLEAN;
BEGIN
  IF a.fleet <> b.fleet THEN
    FlakListLess := a.fleet < b.fleet
  ELSE
    FlakListLess := a.ship_rel < b.ship_rel;
END; { FlakListLess }

PROCEDURE PanelSet(side : INTEGER; ship : INTEGER);
VAR other : INTEGER;
BEGIN
  other := side XOR 1;
  IF NOT vis_panel THEN BEGIN
    vis_panel := TRUE;
    vis_show[0] := ship;
    vis_show[1] := ship;
  END ELSE IF ship <> vis_show[side] THEN BEGIN
    vis_show[side] := ship;
  END ELSE IF ship <> vis_show[other] THEN BEGIN
    vis_show[other] := ship;
  END ELSE BEGIN
    vis_panel := FALSE;
  END;
END;

{ FLAK Visualizer - Main Entry }
PROCEDURE FlakPlayVis(auto_figure_result:BOOLEAN);
VAR
  y_org          : INTEGER;
  time           : LONGINT;
  speed          : INTEGER;
  i              : INTEGER;
  w              : WORD;
  redraw         : (rNone, rStatus, rAll);
  camera         : BOOLEAN;
  l              : LONGINT;
  pending_redraw : BOOLEAN;
  list_top       : FlakListPos;
  list_cur       : FlakListPos;
  list_tmp       : FlakListPos;

CONST list_num_lines = 22;

  PROCEDURE DrawStatusPanel(y : INTEGER);
  CONST x_b1    = 480 + 5;
        x_b2    = 480 + 5 + 16;
  VAR m : INTEGER;
  BEGIN
    DrawButton(x_b1, y + (475-1*16), x_b1 + 30, y + (473-0*16), 'ESC', 0);
    DrawButton(x_b1, y + (475-2*16), x_b1 + 14, y + (473-1*16), 'A', 0);
    DrawButton(x_b2, y + (475-2*16), x_b2 + 14, y + (473-1*16), 'Z', 0);
    IF camera THEN m := bf_Highlight ELSE m := 0;
    DrawButton(x_b1, y + (475-3*16), x_b1 + 14, y + (473-2*16), 'C', m);
    IF vis_mode = vis_Iso THEN m := bf_Highlight ELSE m := 0;
    DrawButton(x_b1, y + (475-4*16), x_b1 + 14, y + (473-3*16), 'I', m);
    IF vis_mode = vis_3D THEN m := bf_Highlight ELSE m := 0;
    DrawButton(x_b2, y + (475-4*16), x_b2 + 14, y + (473-3*16), '3', m);
    DrawButton(x_b1, y + (475-5*16), x_b1 + 14, y + (473-4*16), '+', 0);
    DrawButton(x_b2, y + (475-5*16), x_b2 + 14, y + (473-4*16), '-', 0);
    IF vis_status = vs_Playing THEN m := bf_Highlight ELSE
     IF vis_status >= vs_Terminated THEN m := ws_Disabled ELSE m := 0;
    DrawButton(x_b1, y + (475-6*16), x_b1 + 14, y + (473-5*16), 'P', m);
  END; { DrawStatusPanel }

  PROCEDURE DrawShipList(y : INTEGER);
  VAR p, prev : FlakListPos;
      i, ci : INTEGER;
      s : STRING[63];
      sid : INTEGER;
      eol : BOOLEAN;
  BEGIN
    p := list_top;
    Inc(y, 20);
    TextStyle := 3;
    JustX := 0;
    JustY := 0;
    eol := FALSE;
    FOR i := 1 TO list_num_lines DO BEGIN
      IF eol THEN BEGIN
        s := '';
        AABg := Gray;
        Color := 0;
      END ELSE BEGIN
        IF p.ship_rel = -1 THEN BEGIN
          s := ' Fleet ' + itoa(p.fleet+1) + ': ' + RaceNames[flak_fleets^[p.fleet]^.data.player].Adj;
          IF p.fleet = t_follow_fleet THEN s[1] := #7;
          IF flak_fleets^[p.fleet]^.alive THEN ci := 0 ELSE ci := 2;
        END ELSE BEGIN
          sid := flak_fleets^[p.fleet]^.data.first_ship + p.ship_rel;
          WITH flak_ships^[sid].fs^ DO BEGIN
            IF data.flags AND flak_IsPlanet<>0 THEN s := '  Planet ' + GetStr(data.name, 20)
            ELSE BEGIN
              IF (data.hull > 0) AND (data.hull <= HullCnt) THEN
                s := '  ' + HullsX^[data.hull].ShortName
              ELSE
                s := '  Ship';
              s := s + ' ' + GetStr(data.name, 20);
            END;
            IF status.is_alive THEN ci := 0 ELSE ci := 2;
          END;
        END;
        IF (p.fleet = list_cur.fleet) AND (p.ship_rel = list_cur.ship_rel) THEN BEGIN
          AABg := 0;
          Color := tech_high_colors[ci];
        END ELSE BEGIN
          AABg := Gray;
          Color := tech_low_colors[ci];
        END;
        prev := p;
        FlakListInc(p, 1);
        IF p.ship_rel = prev.ship_rel THEN eol := TRUE;
      END;
      OutFTextXY(485, y, 635 - 485, s);
      Inc(y, 16);
    END;
    AABg := -1;
    TextStyle := 0;
  END;

LABEL TogglePlay;
BEGIN
  y_org := 0;
  speed := 1;
  camera := TRUE;
  vis_inhibit := FALSE;
  list_top.fleet := 0;
  list_top.ship_rel := -1;
  list_cur.fleet := 0;
  list_cur.ship_rel := -1;
  VisInit;

  FullScreen;
  ExtdBar(0, 0, 639, 479, 0);
  IF _double THEN ExtdBar(0, 480, 639, 479+480, 0);

  SetWindow(0, 0, 479, 479);
  VisDrawBattlefield;
  FullScreen;
  VisDrawStatusPanel(0);
  DrawStatusPanel(0);
  DrawShipList(0);
  IF _double THEN BEGIN
    VisDrawStatusPanel(480);
    DrawStatusPanel(480);
    DrawShipList(480);
  END;
  time := OsTimerValue;
  redraw := rNone;
  pending_redraw := FALSE;

  WHILE TRUE DO BEGIN
    list_tmp := list_cur;
    FlakListInc(list_tmp, 1-list_num_lines);
    IF FlakListLess(list_top, list_tmp) THEN list_top := list_tmp;
    IF FlakListLess(list_cur, list_top) THEN list_top := list_cur;
    IF OsTimerValue >= time + speed THEN BEGIN
      time := OsTimerValue;
      IF (vis_status = vs_Playing) OR pending_redraw THEN BEGIN
        { redraw complete battlefield on other page, flip pages }
        IF _double THEN y_org := y_org XOR 480;
        SetWindow(0, y_org, 479, y_org + 479);
        SetOrigin(0, y_org);
        WaitDark;
        ExtdBar(0, 0, 479, 479, 0);
        VisDrawBattlefield;
        IF _double THEN SetStartY(y_org);
        pending_redraw := FALSE;
        IF redraw = rAll THEN redraw := rStatus;
      END;
      IF vis_status = vs_Playing THEN BEGIN
        IF FlakPlayOneTick THEN BEGIN
          vis_status := vs_Terminated;
          IF flak_header_time < 0 THEN flak_header_time := flak_time;
          IF auto_figure_result THEN FlakFigureResults;
        END;
        IF camera THEN BEGIN
          l := VisComputeRange;
          IF l < vis_max_r - 200 THEN Dec(vis_max_r, 200) ELSE
          IF l > vis_max_r + 200 THEN Inc(vis_max_r, 200);
          IF vis_mode = vis_3D THEN CamFollow;
        END;
        pending_redraw := TRUE;
        redraw := rStatus;
      END;
      FullScreen;
      SetOrigin(0, 0);
    END;

    { FIXME: I do not particularily like it that the redraw code
      appears here, again }
    IF redraw >= rAll THEN BEGIN
      { redraw requested }
      IF _double THEN y_org := y_org XOR 480;
      SetWindow(0, y_org, 479, y_org + 479);
      SetOrigin(0, y_org);
      WaitDark;
      ExtdBar(0, 0, 479, 479, 0);
      VisDrawBattlefield;
      IF _double THEN SetStartY(y_org);
      FullScreen;
      SetOrigin(0,0);
      pending_redraw := FALSE;
    END;
    IF redraw >= rStatus THEN BEGIN
      DrawStatusPanel(0);
      DrawShipList(0);
      IF _double THEN BEGIN
        DrawStatusPanel(480);
        DrawShipList(480);
      END;
    END;
    redraw := rNone;

    w := WORD(GetAKey);
    CASE w OF
      { play/pause }
      Ord(' '), Ord('p'): TogglePlay: BEGIN
        IF vis_status = vs_Halt THEN vis_status := vs_Playing ELSE
        IF vis_status = vs_Playing THEN vis_status := vs_Halt;
        redraw := rStatus;
      END;
      { faster }
      Ord('+') : IF speed > 0 THEN BEGIN
        Dec(speed);
        IF vis_status = vs_Halt THEN vis_status := vs_Playing;
        redraw := rStatus;
      END;
      { slower }
      Ord('-') : IF speed < 20 THEN BEGIN
        Inc(speed);
        IF vis_status = vs_Halt THEN vis_status := vs_Playing;
        redraw := rStatus;
      END;
      { toggle panel }
      Ord('1'), Ord('2') : BEGIN
        IF (list_cur.ship_rel >= 0) THEN
          PanelSet(w - Ord('1'), list_cur.ship_rel + flak_fleets^[list_cur.fleet]^.data.first_ship)
        ELSE
          PanelSet(w - Ord('1'), flak_fleets^[list_cur.fleet]^.data.first_ship);
        redraw := rAll;
      END;
      kbDel: IF vis_panel THEN BEGIN
        vis_panel := FALSE;
        redraw := rAll;
      END;
      { toggle 3D }
      Ord('3') : BEGIN
        IF vis_mode = vis_3D THEN vis_mode := vis_Flat ELSE vis_mode := vis_3D;
        redraw := rAll;
      END;
      { toggle Iso 3D }
      Ord('i'): BEGIN
        IF vis_mode = vis_Iso THEN vis_mode := vis_Flat ELSE vis_mode := vis_Iso;
        redraw := rAll;
      END;
      { toggle grid }
      Ord('g'): BEGIN
        vis_grid := NOT vis_grid;
        redraw := rAll;
      END;
      { zoom in }
      Ord('a'): BEGIN
        camera := FALSE;
        IF vis_max_r > 10000 THEN Dec(vis_max_r, 1000);
        redraw := rAll;
      END;
      { zoom away }
      Ord('y'), Ord('z'): BEGIN
        camera := FALSE;
        IF vis_max_r < 2000000 THEN Inc(vis_max_r, 1000);
        redraw := rAll;
      END;
      { camera on/off }
      Ord('c'): BEGIN
        camera := NOT camera;
        redraw := rStatus;
      END;
      { follow }
      Ord('f'): BEGIN
        t_follow_fleet := list_cur.fleet;
        redraw := rAll;
      END;
      { revive }
      Ord('r'): BEGIN
        IF vis_status = vs_Terminated THEN BEGIN
          FOR i := 0 TO flak_ship_cnt-1 DO
           IF flak_ships^[i].fs^.data.ending_status > 0 THEN BEGIN
             flak_ships^[i].fs^.status.is_alive := TRUE;
             flak_fleets^[flak_ships^[i].fs^.fleet_nr]^.alive := TRUE;
           END;
          vis_status := vs_Revived;
          redraw := rAll;
        END;
      END;
      { fast forward }
      kbAltRight, kbCtrlRight: BEGIN
        IF vis_status <= vs_Terminated THEN BEGIN
          vis_inhibit := TRUE;
          Window(190,280+y_org,290,310+y_org);
          JustX:=1;
          JustY:=1;
          Color:=15;
          AAColor:=Gray;
          OutTextXY(240,295+y_org,'Wait...');
          WHILE NOT FlakPlayOneTick DO
            ;
          vis_inhibit := FALSE;
          vis_status := vs_Terminated;
          IF flak_header_time < 0 THEN flak_header_time := flak_time;
          IF auto_figure_result THEN FlakFigureResults;
          redraw := rAll;
        END;
      END;
      { down to next alive ship }
      kbCtrlDown, kbCtrlUp: BEGIN
        list_tmp := list_cur;
        REPEAT
          i := list_tmp.ship_rel;
          IF w=kbCtrlDown THEN FlakListInc(list_tmp, +1)
                          ELSE FlakListInc(list_tmp, -1);
          IF ((list_tmp.ship_rel >= 0)
              AND (flak_ships^[list_tmp.ship_rel + flak_fleets^[list_tmp.fleet]^.data.first_ship].fs^.status.is_alive))
            OR ((list_tmp.ship_rel = -1)
                AND (flak_fleets^[list_tmp.fleet]^.alive))
          THEN BEGIN
            list_cur := list_tmp;
            Break;
          END;
        UNTIL i = list_tmp.ship_rel;
        redraw := rAll;
      END;
      { rewind }
      kbAltLeft, kbCtrlLeft: BEGIN
        vis_status := vs_Halt;
        FlakBattleDone;
        FlakBattleInit;
        redraw := rAll;
      END;
      { camera movement }
      $4800: IF (vis_mode = vis_3D) AND (OsShiftKeys AND ss_Shift <> 0) THEN BEGIN
        camera := FALSE;
        t_cam_hei := t_cam_hei + 0.03;
        redraw := rAll;
      END ELSE BEGIN
        FlakListInc(list_cur, -1);
        redraw := rStatus;
      END;
      $5000: IF (vis_mode = vis_3D) AND (OsShiftKeys AND ss_Shift <> 0) THEN BEGIN
        camera := FALSE;
        t_cam_hei := t_cam_hei - 0.03;
        redraw := rAll;
      END ELSE BEGIN
        FlakListInc(list_cur, +1);
        redraw := rStatus;
      END;
      $4B00: IF (vis_mode = vis_3D) AND (OsShiftKeys AND ss_Shift <> 0) THEN BEGIN
        camera := FALSE;
        t_cam_azi := t_cam_azi + 0.03;
        redraw := rAll;
      END;
      $4D00: IF (vis_mode = vis_3D) AND (OsShiftKeys AND ss_Shift <> 0) THEN BEGIN
        camera := FALSE;
        t_cam_azi := t_cam_azi - 0.03;
        redraw := rAll;
      END ELSE BEGIN
        GOTO TogglePlay;
      END;
      kbPgUp: BEGIN
        FlakListInc(list_cur, 1-list_num_lines);
        redraw := rStatus;
      END;
      kbPgDn: BEGIN
        FlakListInc(list_cur, list_num_lines-1);
        redraw := rStatus;
      END;
      kbHome: BEGIN
        list_cur.fleet := 0;
        list_cur.ship_rel := -1;
        redraw := rStatus;
      END;
      kbEnd: BEGIN
        list_cur.fleet := flak_fleet_cnt-1;
        list_cur.ship_rel := flak_fleets^[flak_fleet_cnt-1]^.data.num_ships-1;
        redraw := rStatus;
      END;
      27: Break;
    END;
  END;

  IF _double THEN SetStartY(0);
  DrawUI;
END;

{=============================== SIMULATION SETUP =======================}

FUNCTION IsAttacking(me, them : INTEGER) : BOOLEAN;
CONST _IgnoredFCodes='???'#0'NTP'#0'mkt'#0'lfm'#0'ATT'#0'NUK';
      IgnoredFCodes:STRING[Length(_IgnoredFCodes)]=_IgnoredFCodes;
VAR meo, themo : PSimObject;
    mes, thems : FlakShipPtr;
    tc, bc     : INTEGER;
    s1         : STRING[3];
LABEL Match;
BEGIN
  mes := flak_ships^[me].fs;
  meo := flak_ships^[me].aux;
  thems := flak_ships^[them].fs;
  themo := flak_ships^[them].aux;

  IsAttacking:=FALSE;
  { same race doesn't attack }
  IF (mes^.data.player = thems^.data.player) THEN Exit;
  { if we are not aggressive or have no fuel, we don't attack }
  IF (meo^.aggressivity = 0) OR (meo^.aggressivity = agr_NoFuel) THEN Exit;
  { if we don't have matching PE, we don't attack }
  IF (meo^.Aggressivity>0) THEN
    IF (meo^.Aggressivity<>themo^.o.Owner) AND (EnemyMatrix[meo^.o.Owner] AND (1 SHL themo^.o.Owner)=0) THEN Exit;
  { a planet can't attack certain ships }
  IF (mes^.data.flags AND flak_IsPlanet<>0) THEN BEGIN
    IF ((RaceIds[thems^.data.player] = 4) AND (HConfigInt(co_FascistAttacked)=0))
     OR ((RaceIds[thems^.data.player] = 10) AND (HConfigInt(co_RebelAttacked)=0))
     {OR ((themo^.Hull>0) AND (HullsX^[themo^.Hull].Text<>NIL)
         AND (Pos('IMPERIAL ASSAULT', Upstr(HullsX^[themo^.Hull].Text^))<>0))}
     { FIXME: query real ship, query for SPC_PlanetImmunity }
      OR HullDoes(themo^.Hull, SPC_ImperialAssault, -1, -1, 1 SHL thems^.data.player, 1 SHL thems^.data.level, 0)
     THEN Exit;
  END;
  { friends don't attack }
  IF Odd(AllianceMatrix[mes^.data.player] SHR thems^.data.player) THEN Exit;
  { can we attack? }
  IF NOT CanStillAttack(mes^, thems^) THEN Exit;

  S1:=GetStr(meo^.FCode,3);
  IF (themo^.aggressivity = agr_NoFuel) THEN BEGIN
    { zero-fuel things are only attacked by planets with NUK,
      but not if they're bird and capital }
    IF mes^.data.flags AND flak_IsPlanet=0 THEN Exit;
    IF s1<>'NUK' THEN Exit;
    IF (RaceIds[thems^.data.player]=3)
      AND ((thems^.data.beam_count>0)
           OR (thems^.data.torp_lcount>0)
           OR (thems^.data.bay_count>0))
      THEN Exit;
  END;
  IF (S1=GetStr(themo^.FCode,3)) THEN BEGIN
    S1 := S1 + '   ';
    IF (xtrfcodes<>NIL) AND IsXtrFcode(S1) THEN GOTO Match;
    IF (Pos(S1, IgnoredFCodes)=0) THEN Exit;
  END;
 Match:
  IsAttacking:=TRUE;
END;

FUNCTION IsUnarmed(p : FlakShipPtr):BOOLEAN;
BEGIN
  IsUnarmed := (p^.data.beam_count=0) AND (p^.data.torp_lcount=0) AND (p^.data.bay_count=0);
END;

FUNCTION ComputeAttackList(VAR f : FlakFleet) : BOOLEAN;
VAR first_entry      : INTEGER;
  possible_opponent  : INTEGER;
  can_attack         : BOOLEAN;
  match_pe, match_fc : BOOLEAN;
  fleet_index        : INTEGER;
  fleet_member       : INTEGER;
  bonus              : INTEGER;
BEGIN
  first_entry := att_list_cnt;

  FOR possible_opponent := 0 TO flak_ship_cnt-1 DO BEGIN
    can_attack := FALSE;
    match_pe := FALSE;
    match_fc := FALSE;
    FOR fleet_index := 1 TO f.data.num_ships DO BEGIN
      fleet_member := f.data.first_ship + fleet_index-1;
      IF IsAttacking(fleet_member, possible_opponent) OR IsAttacking(possible_opponent, fleet_member)
      THEN BEGIN
        IF NOT (IsUnarmed(flak_ships^[fleet_member].fs) AND IsUnarmed(flak_ships^[possible_opponent].fs)) THEN BEGIN
          can_attack := TRUE;
          WITH flak_ships^[fleet_member] DO
            IF (fs^.data.flags AND flak_IsPlanet = 0)
             AND (aux^.aggressivity = flak_ships^[possible_opponent].fs^.data.player)
            THEN match_pe := TRUE;
        END;
      END ELSE BEGIN
        match_fc := TRUE;
      END;
    END;
    IF can_attack THEN BEGIN
      bonus := FlakRandom(FLAK_CONFIG.RatingRandomBonus);
      IF NOT match_pe THEN Inc(bonus, FLAK_CONFIG.RatingPEBonus);
      IF NOT match_fc THEN Inc(bonus, FLAK_CONFIG.RatingFullAttackBonus);
      IF NOT FlakAddAttListEntry(possible_opponent, bonus) THEN BEGIN
        ComputeAttackList := FALSE;
        Exit;
      END;
    END;
  END;

  f.data.att_list_pointer := first_entry;
  f.data.att_list_size    := att_list_cnt - first_entry;
  ComputeAttackList := TRUE;
END;

FUNCTION RemovePassiveObjects : BOOLEAN;
VAR p      : PWordArray;
  i, n     : INTEGER;
  ship_id  : INTEGER;
  fleet_id : INTEGER;
BEGIN
  p := Malloc(Sizeof(WORD) * flak_ship_cnt);
  IF p=NIL THEN BEGIN
    flak_oom := TRUE;
    RemovePassiveObjects := FALSE;
    Exit;
  END;

  { first, mark all referenced ships. We use the fact that a ship
    which is in no attack list can itself not attack anyone and
    therefore is useless. After this loop, all units which are used
    have a 1 in p }
  FOR i:=0 TO flak_fleet_cnt-1 DO BEGIN
    WITH flak_fleets^[i]^ DO BEGIN
      FOR n:=0 TO data.att_list_size-1 DO
        p^[att_list^[data.att_list_pointer + n].ship_nr] := 1;
      {FOR n:=1 TO data.num_ships DO
        flak_ships^[n + data.first_ship-1].fs^.fleet_nr := i;}
    END;
  END;

  { now, recompute stuff }
  ship_id := 0;
  FOR i:=0 TO flak_ship_cnt-1 DO BEGIN
    IF p^[i]=0 THEN BEGIN
      { ship does not appear in lists }
      p^[i] := WORD(-1);
      Dec(flak_fleets^[flak_ships^[i].fs^.fleet_nr]^.data.num_ships);
      FreeMem(flak_ships^[i].fs, Sizeof(FlakShip));
    END ELSE BEGIN
      p^[i] := ship_id;
      flak_ships^[ship_id].fs := flak_ships^[i].fs;
      flak_ships^[ship_id].aux := flak_ships^[i].aux;
      Inc(ship_id);
    END;
  END;
  FOR i:=ship_id TO flak_ship_cnt-1 DO
    flak_ships^[i].fs := NIL;
  flak_ship_cnt := ship_id;

  { remap data: attack lists }
  FOR i:=0 TO att_list_cnt-1 DO
    att_list^[i].ship_nr := p^[att_list^[i].ship_nr];

  { remap data: fleet pointers }
  fleet_id := 0;
  FOR i:=0 TO flak_fleet_cnt-1 DO BEGIN
    IF flak_fleets^[i]^.data.num_ships=0 THEN BEGIN
      { fleet is gone }
      FreeMem(flak_fleets^[i], Sizeof(FlakFleet));
    END ELSE BEGIN
      { fleet still there, figure out new first ship (count
        is already set up). }
      ship_id := flak_fleets^[i]^.data.first_ship;
      WHILE INTEGER(p^[ship_id]) < 0 DO Inc(ship_id);
      flak_fleets^[i]^.data.first_ship := p^[ship_id];
      flak_fleets^[fleet_id] := flak_fleets^[i];
      Inc(fleet_id);
    END;
  END;
  FOR i:=fleet_id TO flak_fleet_cnt-1 DO
    flak_fleets^[i] := NIL;
  flak_fleet_cnt := fleet_id;

  FreeMem(p, Sizeof(WORD) * flak_ship_cnt);
END;

{ Compute initial positions for all fleets }
PROCEDURE ComputeInitialPositions;

  PROCEDURE AssignInitialPositions(player:INTEGER; angle:REAL; dist:LONGINT);
  VAR offset, i : INTEGER;
    this_fleet  : FlakFleetPtr;
  BEGIN
    offset := FLAK_CONFIG.StartingDistancePerFleet;
    FOR i:=0 TO flak_fleet_cnt-1 DO BEGIN
      this_fleet := flak_fleets^[i];
      IF this_fleet^.data.player <> player THEN Continue;
      IF this_fleet^.data.speed > offset THEN
        offset := this_fleet^.data.speed;

      this_fleet^.data.x_init := Round(Cos(angle) * dist);
      this_fleet^.data.y_init := Round(Sin(angle) * dist);
      Inc(dist, offset);
      angle := angle + PI/180;
    END;
  END;

VAR
  planet_owner : INTEGER;
  player_set   : INTEGER;
  num_players  : INTEGER;
  i, fl        : INTEGER;
  counter      : INTEGER;
  offs         : LONGINT;
BEGIN
  planet_owner := 0;
  num_players := 0;
  player_set := 0;
  FOR i:=0 TO flak_ship_cnt-1 DO BEGIN
    WITH flak_ships^[i].fs^ DO BEGIN
      IF data.flags AND flak_IsPlanet<>0 THEN planet_owner := data.player;
      IF (player_set AND (1 SHL data.player)) = 0 THEN BEGIN
        player_set := player_set OR (1 SHL data.player);
        Inc(num_players);
      END;
    END;
  END;

  { assign positions }
  counter := 0;
  offs := FLAK_CONFIG.StartingDistancePerPlayer * LONGINT(num_players);
  FOR fl := 0 TO flak_fleet_cnt-1 DO BEGIN
    i := flak_fleets^[fl]^.data.player;
    IF (player_set AND (1 SHL i)) <> 0 THEN BEGIN
      Dec(player_set, (1 SHL i));
      IF (planet_owner <> 0) THEN BEGIN
        IF i = planet_owner THEN BEGIN
          AssignInitialPositions(i, 0, FLAK_CONFIG.StartingDistancePlanet + offs);
        END ELSE BEGIN
          AssignInitialPositions(i, PI/2 + (2*counter+1) * (PI/2) / (num_players-1),
                                 FLAK_CONFIG.StartingDistanceShip + offs);
          Inc(counter);
        END;
      END ELSE BEGIN
        AssignInitialPositions(i, PI/2 + (2*counter+1) * PI / num_players,
                               FLAK_CONFIG.StartingDistanceShip + offs);
        Inc(counter);
      END;
    END;
  END;
END; { ComputeInitialPositions }

{ Compute fleet speeds. }
PROCEDURE ComputeFleetSpeeds;
VAR i, speed, j, this_speed : INTEGER;
    this_ship : FlakShipPtr;
BEGIN
  FOR i := 0 TO flak_fleet_cnt-1 DO BEGIN
    speed := 0;
    FOR j := 1 TO flak_fleets^[i]^.data.num_ships DO BEGIN
      this_ship := flak_ships^[flak_fleets^[i]^.data.first_ship + j - 1].fs;
      this_ship^.fleet_nr := i;
      IF this_ship^.data.flags AND flak_IsPlanet<>0 THEN
        this_speed := 0
      ELSE IF this_ship^.data.player > 11 THEN
        this_speed := pbattleconf^[11].ShipMovementSpeed
      ELSE
        this_speed := pbattleconf^[this_ship^.data.player].ShipMovementSpeed;
      IF (j=1) OR (this_speed < speed) THEN speed := this_speed;
    END;
    flak_fleets^[i]^.data.speed := speed;
  END;
END; { ComputeFleetSpeeds }

{ Initialize simulation.
  Precondition: you have called
  - FlakInit to allocate memory
  - FlakAdd(...) to add all the ships, including the aux pointers
  Postcondition: you can call PlayOneTick etc to play it }
FUNCTION FlakInitSimulation : BOOLEAN;
VAR random_perm : ARRAY[1..FLAK_MAX_PLAYERS] OF INTEGER;

  { returns normal FCBO value plus 100 }
  FUNCTION ShipFCBOValue(ship : INTEGER) : INTEGER;
  VAR fcbo : INTEGER;
      fc   : TFCode;
      j : INTEGER;
  BEGIN
    WITH flak_ships^[ship] DO BEGIN
      fcbo := 1099;
      fc   := aux^.FCode;
      IF (fs^.data.flags AND flak_IsPlanet)<>0 THEN BEGIN
        IF (fc = 'ATT') OR (fc = 'NUK') THEN fcbo := 100;
      END;
      IF (fc[2]>='0') AND (fc[2]<='9') AND (fc>='0') AND (fc<='9') THEN BEGIN
        j := 10*(Ord(fc[2])-48) + Ord(fc[3])-48;
        IF (fc[1]='-') THEN fcbo := 100-j ELSE
        IF (fc[1]>='0') AND (fc[1]<='9') THEN fcbo := 100+100*(Ord(fc[1])-48) + j;
      END;
    END;
    IF fcbo > 1099 THEN fcbo := 1099;
    ShipFCBOValue := fcbo;
  END;

  FUNCTION IsShipBefore(ship, other : INTEGER) : BOOLEAN;
  VAR ks, ko : INTEGER;
  BEGIN
    { first, sort by players }
    ks := random_perm[flak_ships^[ship].fs^.data.player];
    ko := random_perm[flak_ships^[other].fs^.data.player];
    IF ks<>ko THEN BEGIN
      IsShipBefore := ks < ko;
      Exit;
    END;

    { next, compare FCBO }
    ks := ShipFCBOValue(ship);
    ko := ShipFCBOValue(other);
    IF ks<>ko THEN BEGIN
      IsShipBefore := ks < ko;
      Exit;
    END;

    { now, test Ids }
    ks := flak_ships^[ship].fs^.data.id;
    ko := flak_ships^[other].fs^.data.id;
    IF ks<>ko THEN BEGIN
      IsShipBefore := ks < ko;
      Exit;
    END;

    { finally, if ship is a planet, it goes first }
    IsShipBefore := (flak_ships^[ship].fs^.data.flags AND flak_IsPlanet)<>0;
  END; { IsShipBefore }

VAR i, j, k                                         : INTEGER;
    a                                               : FlakShipPtr;
    b                                               : PSimObject;
    player_count, player_set                        : INTEGER;
    cur_player, cur_fcbo, cur_fleet, cur_fleet_size : INTEGER;
    cur_planet                                      : INTEGER;

BEGIN
  FlakInitSimulation := FALSE;

  { sanity checks }
  IF flak_ship_cnt < 2 THEN Exit;
  FOR i:=0 TO flak_ship_cnt-1 DO BEGIN
    WITH flak_ships^[i] DO BEGIN
      IF aux=NIL THEN Exit;
      IF (fs^.data.player <= 0) OR (fs^.data.player > FLAK_MAX_PLAYERS) THEN Exit;
    END;
  END;

  { first, generate a random permutation of players }
  FOR i:=1 TO FLAK_MAX_PLAYERS DO random_perm[i] := i;
  FOR i:=2 TO FLAK_MAX_PLAYERS DO BEGIN
    j := FlakRandom(i) + 1;
    k := random_perm[j];
    random_perm[j] := random_perm[i];
    random_perm[i] := k;
  END;

  { now sort ships. Standard bubblesort, because I'm lazy }
  i := 0;
  WHILE i < flak_ship_cnt-1 DO BEGIN
    IF IsShipBefore(i+1, i) THEN BEGIN
      a := flak_ships^[i].fs;
      b := flak_ships^[i].aux;
      flak_ships^[i].fs := flak_ships^[i+1].fs;
      flak_ships^[i].aux := flak_ships^[i+1].aux;
      flak_ships^[i+1].fs := a;
      flak_ships^[i+1].aux := b;
      IF i > 0 THEN Dec(i);
    END ELSE Inc(i);
  END;

  { count players }
  { FIXME: I guess this can safely be left out }
  player_count := 0;
  player_set := 0;
  FOR i:=0 TO flak_ship_cnt-1 DO BEGIN
    j := flak_ships^[i].fs^.data.player;
    IF (player_set AND (1 SHL j)) = 0 THEN BEGIN
      player_set := player_set OR (1 SHL j);
      Inc(player_count);
    END;
  END;
  IF player_count < 2 THEN Exit;

  { build fleets }
  cur_player := 0;
  cur_fcbo := 0;
  cur_planet := 0;
  cur_fleet := 0;
  cur_fleet_size := 0;
  FOR i:=0 TO flak_ship_cnt-1 DO BEGIN
    IF (i=0) OR (cur_player<>flak_ships^[i].fs^.data.player)
      OR (cur_fcbo <> ShipFCBOValue(i) DIV 100)
      OR (cur_planet <> (flak_ships^[i].fs^.data.flags AND flak_IsPlanet))
      OR (cur_fleet_size >= FLAK_CONFIG.MaximumFleetSize)
    THEN BEGIN
      IF NOT FlakFleetAdd THEN Exit;
      cur_fleet := flak_fleet_cnt-1;
      cur_fcbo := ShipFCBOValue(i) DIV 100;
      cur_planet := flak_ships^[i].fs^.data.flags AND flak_IsPlanet;
      cur_player := flak_ships^[i].fs^.data.player;
      cur_fleet_size := 0;
      WITH flak_fleets^[cur_fleet]^ DO BEGIN
        data.player := cur_player;
        data.first_ship := i;
      END;
    END;
    Inc(flak_fleets^[cur_fleet]^.data.num_ships);
    Inc(cur_fleet_size);
    { initialize ship. ComputeAttackList wants to call CanStillAttack which
      needs .status initialized; RemovePassiveObjects needs .fleet_nr }
    FlakShipInit(flak_ships^[i].fs^, i, cur_fleet);
  END;

  { compute attack lists }
  FOR i:=0 TO flak_fleet_cnt-1 DO
    IF NOT ComputeAttackList(flak_fleets^[i]^) THEN Exit;

  { preparations }
  IF NOT RemovePassiveObjects THEN Exit;
  ComputeFleetSpeeds;
  ComputeInitialPositions;

  FlakInitSimulation := TRUE;
END; { FlakInitSimulation }

PROCEDURE FlakInitCommon(VAR item : TFlakShip);
BEGIN
  item.max_fighters_launched := MFL_SCALE * item.bay_count;
  IF item.max_fighters_launched < MINIMUM_MFL THEN item.max_fighters_launched := MINIMUM_MFL;
  IF item.max_fighters_launched > MAXIMUM_MFL THEN item.max_fighters_launched := MAXIMUM_MFL;
  item.rating := (item.mass * LONGINT(FLAK_CONFIG.RatingMassScale)
                + item.torp_lcount * item.torp_type * LONGINT(FLAK_CONFIG.RatingTorpScale)
                + item.beam_count * item.beam_type * LONGINT(FLAK_CONFIG.RatingBeamScale)
                + item.bay_count * LONGINT(FLAK_CONFIG.RatingBayScale));
  item.compensation_rating := (FLAK_CONFIG.CompensationShipScale
                               + item.torp_lcount * LONGINT(FLAK_CONFIG.CompensationTorpScale)
                               + item.beam_count * LONGINT(FLAK_CONFIG.CompensationBeamScale)
                               + item.bay_count * LONGINT(FLAK_CONFIG.CompensationFighterScale));
  IF item.compensation_rating > FLAK_CONFIG.CompensationLimit
    THEN item.compensation_rating := FLAK_CONFIG.CompensationLimit;
  item.ending_status := 0;
END; { FlakInitCommon }

{ Find captor for /ship_nr/. Returns ship number or -1 }
FUNCTION FindCaptor(ship_nr : INTEGER) : INTEGER;
VAR
  last       : INTEGER;
  counts     : ARRAY[1..FLAK_MAX_PLAYERS] OF INTEGER;
  total      : INTEGER;
  i          : INTEGER;
  pick_mask  : INTEGER;
  pick_count : INTEGER;
BEGIN
  { captor still alive? }
  last := flak_ships^[ship_nr].fs^.status.last_hit_by;
  IF (last < 0) OR (flak_ships^[last].fs^.status.is_alive) THEN BEGIN
    FindCaptor := last;
    Exit;
  END;

  { count }
  FillChar(counts, Sizeof(counts), 0);
  total := 0;
  FOR i := 0 TO flak_ship_cnt-1 DO
    IF flak_ships^[i].fs^.status.is_alive THEN BEGIN
      Inc(counts[flak_ships^[i].fs^.data.player]);
      Inc(total);
    END;

  { captor race still alive? }
  i := flak_ships^[last].fs^.data.player;
  IF counts[i] > 0 THEN BEGIN
    pick_mask := 1 SHL i;
    pick_count := counts[i];
  END ELSE BEGIN
    { original owner still alive? }
    i := flak_ships^[ship_nr].fs^.data.player;
    IF counts[i] > 0 THEN BEGIN
      pick_mask := 1 SHL i;
      pick_count := counts[i];
    END ELSE BEGIN
      { anyone else still alive? }
      pick_mask := -1;
      pick_count := total;
    END;
  END;

  { pick one }
  pick_count := FlakRandom(pick_count);
  FOR i := 0 TO flak_ship_cnt-1 DO
    IF (flak_ships^[i].fs^.status.is_alive) AND (pick_mask AND (1 SHL (flak_ships^[i].fs^.data.player)) <> 0)
    THEN BEGIN
      IF pick_count = 0 THEN BEGIN
        FindCaptor := i;
        Exit;
      END;
      Dec(pick_count);
    END;
  FindCaptor := -1;
END;

PROCEDURE FlakFigureResults;
VAR
  i      : INTEGER;
  fs     : FlakShipPtr;
  captor : INTEGER;
  limit  : INTEGER;
BEGIN
  FOR i := 0 TO flak_ship_cnt-1 DO BEGIN
    fs := flak_ships^[i].fs;
    IF fs^.status.is_alive THEN BEGIN
      fs^.data.ending_status := 0;
    END ELSE BEGIN
      captor := FindCaptor(i);
      IF fs^.data.flags AND flak_IsPlanet<>0 THEN BEGIN
        IF captor >= 0 THEN
          fs^.data.ending_status := flak_ships^[captor].fs^.data.player
        ELSE
          fs^.data.ending_status := -1;
      END ELSE BEGIN
        IF (RaceIds[fs^.data.player] = 2)
         AND (captor >= 0) AND (RaceIds[flak_ships^[captor].fs^.data.player] = 2)
          THEN limit := 150 ELSE limit := 99;
        IF (captor >= 0) AND (fs^.status.crew < 0.5)
         AND (Round(fs^.status.damage) <= limit)
          THEN BEGIN
            fs^.data.ending_status := flak_ships^[captor].fs^.data.player
          END ELSE BEGIN
            fs^.data.ending_status := -1;
          END;
      END;
    END;
  END;
END; { FlakFigureResults }

{***************** User Interface *********************}

TYPE TFlakBattleEntry = RECORD
        filepos : LONGINT;
        header  : TFlakBattle;
     END;
     TFlakBattleEntryArray = ARRAY[1..100] OF TFlakBattleEntry;
     PFlakBattleEntryArray = ^TFlakBattleEntryArray;

TYPE PFlakSelector = ^CFlakSelector;
     CFlakSelector = OBJECT(CWidget)
          current, max : INTEGER;
          a            : PFlakBattleEntryArray;
          PROCEDURE Draw; VIRTUAL;
          PROCEDURE SetVal(i:INTEGER);
     END;

PROCEDURE CFlakSelector.Draw;
VAR s : STRING[15];
BEGIN
  Color := 0;
  AAColor := Dark;
  JustX := 0;
  JustY := 0;
  ExtdBar(110,110,400,139,Gray);
  ExtdBar(110,140,500,279,Gray);
  OutTextXY(110,110,'Battle #'+itoa(current)+' of '+itoa(max)+':');
  TextStyle := 3;

  WITH a^[current] DO BEGIN
    OutTextXY(120,140, itoa(header.stuff[iShips].num) + ' units');
    OutTextXY(120,156, itoa(header.stuff[iFleets].num) + ' groups');
    OutTextXY(120,180, 'Location: ' + itoa(header.x) + ',' + itoa(header.y));
    IF header.total_time > 0 THEN BEGIN
      Str(header.total_time MOD 60 : 2, s);
      IF s[1]=' ' THEN s[1] := '0';
      OutTextXY(120,196, 'Time: ' + itoa(header.total_time DIV 60) + ':' + s);
    END;
    OutTextXY(120,212,'Size: ' + itoa(header.this_size) + ' bytes');
  END;

  TextStyle := 0;
END;

PROCEDURE CFlakSelector.SetVal(i:INTEGER);
BEGIN
  IF i <= 0 THEN i := 1;
  IF i > max THEN i := max;
  IF i<>current THEN BEGIN
    current := i;
    DrawWidget;
  END;
END;

FUNCTION Validate(a : PFlakBattleEntryArray; n : INTEGER):BOOLEAN;
VAR i : INTEGER;
BEGIN
  Validate := FALSE;
  FOR i := 1 TO n DO
    WITH a^[i].header DO BEGIN
      IF (this_size <= Sizeof(TFlakBattle))
        OR (stuff[iFleets].size <> Sizeof(TFlakFleet))
        OR (stuff[iShips].size <> Sizeof(TFlakShip))
        OR (stuff[iAttList].size <> 4) THEN Exit;
    END;
  Validate := TRUE;
END; { Validate }

FUNCTION FlakLoadFileEntry(VAR t : TStream; CONST a : TFlakBattleEntry) : BOOLEAN;
VAR
  fs : TFlakShip;
  fl : TFlakFleet;
  ae : ARRAY[1..2] OF INTEGER;
  i  : INTEGER;
BEGIN
  { initialize }
  FlakLoadFileEntry := FALSE;
  flak_oom := FALSE;
  FlakInit(a.header.stuff[iShips].num);
  IF flak_oom THEN Exit;

  { load ships }
  t.Seek(a.filepos + a.header.stuff[iShips].pos);
  FOR i := 1 TO a.header.stuff[iShips].num DO BEGIN
    t.Read(fs, Sizeof(fs));
    IF NOT FlakAdd(fs, NIL) THEN Exit;
  END;

  { FIXME: fleet / attlist loading could be sped up by
    pre-allocating memory }

  { load fleets }
  t.Seek(a.filepos + a.header.stuff[iFleets].pos);
  FOR i := 1 TO a.header.stuff[iFleets].num DO BEGIN
    t.Read(fl, Sizeof(fl));
    IF (fl.first_ship < 0) OR (fl.first_ship >= a.header.stuff[iShips].num)
     OR (fl.num_ships <= 0) OR (fl.num_ships + fl.first_ship > a.header.stuff[iShips].num)
      THEN Exit;
    IF (fl.att_list_pointer < 0) OR (fl.att_list_pointer > a.header.stuff[iAttList].num)
     OR (fl.att_list_size < 0) OR (fl.att_list_size + fl.att_list_pointer > a.header.stuff[iAttList].num)
      THEN Exit;
    IF NOT FlakFleetAdd THEN Exit;
    flak_fleets^[flak_fleet_cnt-1]^.data := fl;
  END;

  { load attack list }
  t.Seek(a.filepos + a.header.stuff[iAttList].pos);
  FOR i := 1 TO a.header.stuff[iAttList].num DO BEGIN
    t.Read(ae, Sizeof(ae));
    IF (ae[1] < 0) OR (ae[1] >= a.header.stuff[iShips].num) THEN Exit;
    IF NOT FlakAddAttListEntry(ae[1], ae[2]) THEN Exit;
  END;

  { setup rest }
  flak_header_seed := a.header.seed;
  flak_header_time := a.header.total_time;

  FlakBattleInit;
  IF flak_oom THEN Exit;

  FlakLoadFileEntry := TRUE;
END;

PROCEDURE FlakPlayFile(VAR t : TStream; CONST s : STRING; current : POINTER);
VAR
  header : TFlakHeader;
  L      : LONGINT;
  a      : PFlakBattleEntryArray;
  i      : INTEGER;
  pw     : PUIWindow;
  pv     : PFlakSelector;
  E      : CEvent;

CONST
  cm_PrevVCR                = 3000;
  cm_NextVCR                = 3001;
  xl : ARRAY[0..12] OF WORD = (kbUp,   cm_PrevVCR, kbDown, cm_NextVCR,
                               kbPgUp, cm_PrevVCR, kbPgDn, cm_NextVCR,
                               kbWheelUp, cm_PrevVCR, kbWheelDown, cm_NextVCR,
                               0);
BEGIN
  IF pbattleconf=NIL THEN BEGIN
    NMessageBox('The file ' + s + ' cannot be played because you do not have the required configuration file.',
                Leerstring);
    Exit;
  END;

  { scan file }
  t.Seek(0);
  t.Read(header, Sizeof(header));
  IF (header.filefmt_version <> 0) THEN BEGIN
    NMessageBox('This file format version is not supported.', Leerstring);
    Exit;
  END;
  IF (header.num_battles = 0) THEN BEGIN
    NMessageBox('The file ' + s + ' does not contain any battles.', Leerstring);
    Exit;
  END;
  L := Sizeof(TFlakBattleEntry) * LONGINT(header.num_battles);
  IF L > 65520 THEN BEGIN
    NMemError;
    Exit;
  END;
  a := Malloc(L);
  IF a = NIL THEN BEGIN
    NMemError;
    Exit;
  END;
  FOR i := 1 TO header.num_battles DO BEGIN
    a^[i].filepos := t.GetPos;
    t.Read(a^[i].header, Sizeof(TFlakBattle));
    t.Seek(a^[i].filepos + a^[i].header.this_size);
  END;
  IF (t.Status<>0) OR NOT Validate(a, header.num_battles) THEN BEGIN
    FreeMem(a, L);
    NMessageBox('The file ' + s + ' is damaged.', Leerstring);
    Exit;
  END;

  pw := New(PUIWindow, Init(100, 80, 540, 365, 'Visual Combat Recorder', wfc_BlueWindow, Gray));

  pw^.Add(New(PButton, InitR(cm_PrevVCR, 110, 330, 25, 25, '', 0, Ord('-'))));
  pw^.Add(New(PButton, InitR(cm_NextVCR, 140, 330, 25, 25, '', 0, Ord('+'))));
  pw^.Add(New(PButton, InitR(cm_Enter, 170, 330, 150, 25, 'ENTER - Play', 0, 13))
                       ^.SetHelp('Play this combat recording'));
  pw^.Add(New(PButton, InitR(cm_Escape, 460, 330, 40, 25, ESCtext, 0, 27))
                       ^.SetHelp(ESChelp));
  pw^.Add(New(PButton, InitR(cm_Help, 505, 330, 25, 25, LetterH, bf_Help, Ord('H'))));
  pv := New(PFlakSelector, Init(110, 110, 530, 300, 0));
  pv^.max := header.num_battles;
  pv^.a := a;
  pv^.current := 1;
  pw^.Add(pv);
  pw^.Add(New(PKeyTrans, Init(@xl)));
  IF current<>NIL THEN pv^.SetVal(INTEGER(current^));
  OpenWindow(pw, TRUE);

  REPEAT
    WinLoop(0, E);
    CASE E.Cmd OF
      cm_PrevVCR: pv^.SetVal(pv^.current - 1);
      cm_NextVCR: pv^.SetVal(pv^.current + 1);
      cm_Enter: BEGIN
        IF NOT FlakLoadFileEntry(t, a^[pv^.current]) THEN BEGIN
          IF flak_oom THEN NMemError ELSE NMessageBox('Unable to load battle.', Leerstring);
        END ELSE BEGIN
          MouseOff;
          vis_inhibit := FALSE;
          FlakPlayVis(TRUE);
        END;
        FlakDone;
      END;
      cm_Help: NHelp(hcSelectVCR);
    END; { case }
  UNTIL E.Cmd = cm_Escape;
  IF current<>NIL THEN INTEGER(current^) := pv^.current;
  CloseDelayed;

  FreeMem(a, L);
END;

END.

{
Local Variables:
mode: pascal
pascal-indent-level: 2
End:
}
